from traits.api import (
    Any,
    Str,
    Bool,
    Int,
    File,
    List,
    Range,
    PrefixList,
    HasRequiredTraits,
)
from Exscript.protocols.exception import InvalidCommandException
from Exscript.protocols.exception import TimeoutException  # <-- expect timeout
from Exscript import Account, PrivateKey
from Exscript.protocols import SSH2
import Exscript

# from paramiko.transport import ConnectionResetError
from paramiko.ssh_exception import SSHException
import paramiko.ssh_exception
import paramiko


from textfsm import TextFSM

from loguru import logger

import arrow

"""Interface down logs: 'show log messages | match SNMP_TRAP_LINK_DOWN'"""

from getpass import getpass, getuser
from io import StringIO
from ipaddress import ip_address as stdlib_ip_factory
import socket as sock
import pkg_resources
import atexit
import time
import json
import copy
import sys
import re
import os

# tests may fail if there are version conflicts...
pkg_resources.require("Exscript==2.6.3")
pkg_resources.require("loguru==0.5.3")
pkg_resources.require("traits==6.3.2")
pkg_resources.require("textfsm==1.1.2")
pkg_resources.require("arrow>=1.2.1")

# Base class for server tests...
# - Exscript/tests/Exscript/protocols/ProtocolTest.py
#
# Test suite for the SSH2 client...
# - Exscript/tests/Exscript/protocols/SSH2Test.py
#
# Test suite for the Dummy client...
# - Exscript/tests/Exscript/protocols/DummyTest.py


## Deprecating these for now...
# from Exscript import Logger as logger_exscript
# from Exscript.util.log import log_to


"""
Copyright 2022 - David Michael Pennington
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1.  Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

2.  Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

3.  Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from
    this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

"""
TODO:

1. Incorporate tests against the Cisco free router testbeds
described by bigevilbeard, below...

- https://github.com/bigevilbeard/Basic_NetMiko/README.md
  - https://devnetsandbox.cisco.com/RM/Diagram/Index/e83cfd31-ade3-4e15-91d6-3118b867a0dd?diagramType=Topology
  - https://devnetsandbox.cisco.com/RM/Diagram/Index/38ded1f0-16ce-43f2-8df5-43a40ebf752e?diagramType=Topology

2. Fix Exscript SSH with legacy ciphers and kexexchange algorithms -> https://github.com/knipknap/exscript/issues/215
"""

HOST = ""
USERNAME = getuser()
PASSWORD = ""
DEFAULT_CONNECT_TIMEOUT = 10
DEFAULT_PROMPT_LIST = []
DEFAULT_PROMPT_TIMEOUT = 30

logger_id = logger.add(sys.stderr)


def valid_path(filepath=""):
    return os.path.isfile(os.path.expanduser(filepath))


@logger.catch(default=True, onerror=lambda _: sys.exit(1))
def account_factory(username=getuser(), password=None, private_key="", keytype="rsa"):
    assert username != ""
    assert keytype in set(["rsa", "dsa", "ecdsa", "ed25519"])

    if password is None:
        password = getpass("Login password for %s: " % username)

    private_key_obj = None
    if isinstance(private_key, str) and private_key != "":
        keypath = os.path.expanduser(private_key)
        assert os.path.isfile(keypath)
        private_key_obj = PrivateKey(keytype=keytype).from_file(keypath)

    account = Account(name=username, password=password, key=private_key_obj)
    return account


# @log_args
@logger.catch(default=True, onerror=lambda _: sys.exit(1))
class Shell(HasRequiredTraits):
    host = Str(value="", required=True)
    port = Range(value=22, low=1, high=65534)
    username = Str(value=getuser(), required=False)
    password = Str(value="", required=False)
    private_key_path = File(value=os.path.expanduser("~/.ssh/id_rsa"))
    inventory = File(value=os.path.expanduser("~/inventory.ini"), required=False)
    # FIXME - needs more drivers... -> https://exscript.readthedocs.io/en/latest/Exscript.protocols.drivers.html
    driver = PrefixList(
        value="generic", values=["generic", "shell", "junos", "ios"], required=False
    )
    termtype = PrefixList(
        value="dumb", values=["dumb", "xterm", "vt100"], required=False
    )
    protocol = PrefixList(value="ssh", values=["ssh"], required=False)
    stdout = PrefixList(value=None, values=[None, sys.stdout], required=False)
    stderr = PrefixList(value=sys.stderr, values=[None, sys.stderr], required=False)
    banner_timeout = Range(value=20, low=1, high=30, required=False)
    connect_timeout = Range(value=10, low=1, high=30, required=False)
    prompt_timeout = Range(value=10, low=1, high=65535, required=False)
    prompt_list = List(Str, required=False)
    default_prompt_list = List(re.Pattern, required=False)
    account = Any(value=None, required=True)
    account_list = List(Exscript.account.Account, required=False)
    json_logfile = File(value="/dev/null", required=False)
    jh = Any(value=None)
    encoding = PrefixList(value="utf-8", values=["latin-1", "utf-8"], required=False)
    interact = Bool(value=False, values=[True, False])
    downgrade_ssh_crypto = Bool(value=False, values=[True, False])
    ssh_attempt_number = Range(value=1, low=1, high=3, required=False)
    conn = Any(required=False)
    debug = Range(value=0, low=0, high=5, required=False)
    allow_invalid_command = Bool(value=True, values=[True, False], required=False)
    MAX_SSH_ATTEMPT = Int(3)

    def __init__(self, **traits):
        super().__init__(**traits)

        assert self.host != ""
        assert isinstance(self.port, int)
        assert len(self.account_list) == 0
        if isinstance(self.account, Account):
            self.append_account(self.account)
        else:
            raise ValueError("Account must be included in the Shell() call")

        # Ensure this was NOT called with username
        if traits.get("username", False) is not False:
            raise ValueError("Shell() calls with username are not supported")

        # Ensure this was NOT called with password
        if traits.get("password", False) is not False:
            raise ValueError("Shell() calls with password are not supported")

        # Check whether host matches an ip address in the inventory...
        host_address = self.search_inventory_for_host_address(self.host)
        print("DISCOVERED HOST in %s" % self.inventory, host_address)

        self.conn = self.do_ssh_login(debug=self.debug)
        self.allow_invalid_command = True

        # Always store the original prompt(s) so we can fallback to them later
        self.default_prompt_list = self.conn.get_prompt()

        # Populate the initial prompt list...
        if self.json_logfile != "/dev/null":
            self.open_json_log()
            self.json_log_entry(cmd="ssh2", action="login", timeout=False)

        if self.interact is True:
            self.interactive_session()

    def __repr__(self):
        return """<Shell: %s>""" % self.host

    def interactive_session(self):
        """Basic method to run an interactive session
        TODO - write timestamped interactive session to disk...
        """

        finished = False
        while finished is not True:
            cmd = input("interact-prompt# ")
            self.execute(cmd, consume=False, timeout=10)
            for line in self.conn.response.splitlines():
                print(line.strip())



    def search_inventory_for_host_address(self, hostname=None):
        """Use an inventory file to map the input hostname to an IPv4 or IPv6 address"""

        # Search inventory first...
        for line in self.iter_inventory_lines():
            # Try to resolve the address as an ansible hostfile...
            mm = re.search(
                r"^\s*({0})\s.*?(ansible_host)\s*=\s*(\S+)".format(hostname.lower()),
                line.lower(),
            )
            if mm is not None:
                host_ip = mm.group(3)
                tmp = stdlib_ip_factory(host_ip)  # This will raise a ValueError
                return host_ip

        else:
            logger.debug("No inventory match for '%s'" % hostname)

        # If this resolves to a valid ipv4 string, return the ipv4 string
        valid_ipv4 = sock.getaddrinfo(hostname, None, sock.AF_INET)[0][4][0]
        logger.debug(valid_ipv4)
        try:
            if stdlib_ip_factory(valid_ipv4):
                return valid_ipv4
        except ValueError:
            logger.debug("No ipv4 DNS record for '%s'" % hostname)
            # Fall through to the next step
            pass

        # If this resolves to a valid ipv6 string, return the ipv6 string
        valid_ipv6 = sock.getaddrinfo(hostname, None, sock.AF_INET6)[0][4][0]
        logger.debug(valid_ipv6)
        try:
            if stdlib_ip_factory(valid_ipv6):
                logger.debug("No ipv4 DNS record for '%s'" % hostname)
                return valid_ipv6
        except ValueError:
            # Fall through to the next step
            pass

        # Raise a non-fatal error...
        logger.warning(
            "Could not resolve or match hostname='%s' in the inventory file: %s"
            % (hostname, self.inventory)
        )
        return None

    def iter_inventory_lines(self):
        if os.path.isfile(os.path.expanduser(self.inventory)):
            with open(self.inventory, "r", encoding="utf=8") as fh:
                for line in fh.read().splitlines():
                    yield line.lower()
        else:
            raise OSError("Cannot find inventory file named '%s'" % self.inventory)

    def do_ssh_login(
        self,
        connect_timeout=10,
        debug=0,
    ):
        assert isinstance(connect_timeout, int)
        assert isinstance(debug, int)
        assert len(self.account_list) > 0

        # FIXME - clean up PrivateKey here...
        private_key = PrivateKey(keytype="rsa").from_file(self.private_key_path)

        self.downgrade_ssh_crypto = False
        if self.protocol == "ssh":

            for self.ssh_attempt_number in [1, 2, 3]:

                assert self.ssh_attempt_number <= self.MAX_SSH_ATTEMPT

                # You have to change allowed ciphers / key exchange options
                # **before** the connection
                #     -> https://stackoverflow.com/a/31303321/667301
                if self.downgrade_ssh_crypto is True:
                    paramiko.Transport._preferred_ciphers = (
                        "aes128-cbc",
                        "3des-cbc",
                    )
                    paramiko.Transport._preferred_kex = (
                        "diffie-hellman-group-exchange-sha1",
                        "diffie-hellman-group14-sha1",
                        "diffie-hellman-group1-sha1",
                    )

                conn = SSH2(driver=self.driver)

                # Save default values...
                DEFAULT_CONNECT_TIMEOUT = conn.get_connect_timeout()
                DEFAULT_PROMPT_LIST = conn.get_prompt()
                DEFAULT_PROMPT_TIMEOUT = conn.get_timeout()

                # FIXME - Exscript should be handling this but the pypi pkg doesn't
                #

                conn.set_connect_timeout(connect_timeout)
                try:
                    conn.connect(hostname=self.host, port=self.port)
                    break

                except sock.timeout as ee:
                    self.downgrade_ssh_crypto = True
                    if self.ssh_attempt_number == self.MAX_SSH_ATTEMPT:
                        error = "Timeout connecting to TCP port {1} on host:{0}".format(
                            self.host, self.port
                        )
                        logger.critical(error)
                        raise OSError(error)
                    else:
                        assert self.ssh_attempt_number < self.MAX_SSH_ATTEMPT
                        time.sleep(0.5)

                except SSHException as ee:
                    self.downgrade_ssh_crypto = True
                    if self.ssh_attempt_number == self.MAX_SSH_ATTEMPT:
                        error = (
                            "Connection to host:{0} on TCP port {1} was reset".format(
                                self.host, self.port
                            )
                        )
                        logger.critical(error)
                        raise OSError(error)
                    else:
                        assert self.ssh_attempt_number < self.MAX_SSH_ATTEMPT
                        time.sleep(0.5)

            login_success = False
            for account in self.account_list:
                conn.login(account)
                try:
                    assert isinstance(conn, SSH2)  # This succeeds if logged in...
                    login_success = True
                    break
                except AssertionError as aa:
                    # login with account failed...
                    continue

            assert login_success is True
            if login_success is True:
                self.password = account.password
            else:
                raise ValueError("Login to host='%s' failed" % self.host)

            conn.set_connect_timeout(DEFAULT_CONNECT_TIMEOUT)

            return conn

        else:
            raise ValueError("FATAL: proto='%s' isn't a valid protocol" % proto)

    ## TODO - we should use a true try / except here...
    def open_json_log(self):
        if self.json_logfile == "/dev/null":
            return None
        else:
            self.jh = open(os.path.expanduser(self.json_logfile), "w", encoding="utf-8")
            atexit.register(self.close_json_log)
            return True

    ## TODO - we should use a true try / except here...
    def close_json_log(self):
        if self.json_logfile == "/dev/null":
            return None
        else:
            self.jh.flush()
            self.jh.close()
            return True

    def json_log_entry(self, cmd=None, action=None, result=None, timeout=False):
        if self.json_logfile == "/dev/null":
            return None
        assert isinstance(cmd, str)
        assert isinstance(action, str)
        assert action in set(["login", "execute", "send", "expect", "output"])
        assert isinstance(result, str) or (result is None)
        assert isinstance(timeout, bool)
        # Pretty json output... or reference json docs
        #     https://stackoverflow.com/a/12944035/667301
        self.jh.write(
            json.dumps(
                {
                    "time": str(arrow.now()),
                    "cmd": cmd,
                    "host": self.host,
                    "action": action,
                    "result": result,
                    "timeout": timeout,
                },
                indent=4,
                sort_keys=True,
            )
            + ","
            + os.linesep
        )

    def append_account(self, account):
        # From the exscript docs...
        #     key = PrivateKey.from_file('~/.ssh/id_rsa', 'my_key_password')
        self.account_list.append(account)

    def _extend_prompt(self, prompt_list=()):
        retval = list()
        for ii in prompt_list:
            if isinstance(ii, str):
                compiled = re.compile(ii)
                retval.append(compiled)

            elif isinstance(ii, re.Pattern):
                retval.append(ii)

            else:
                raise ValueError("Cannot process prompt:'%s'" % ii)

        for ii in self.conn.get_prompt():
            retval.append(ii)

        self.conn.set_prompt(retval)

    def tfsm(self, template=None, input_str=None):
        """Run the textfsm template against input_str"""
        assert isinstance(template, str)
        assert isinstance(input_str, str)
        if os.path.isfile(os.path.expanduser(str(template))):
            # open the textfsm template from disk...
            fh = open(template, "r")
        else:
            # build a fake filehandle around textfsm template string
            fh = StringIO(template)
        fsm = TextFSM(fh)
        header = fsm.header
        values = fsm.ParseText(input_str)
        assert values != [], "Could not match any values with the template."

        ## Pack the extracted values into a list of dicts, using keys from
        ##   the header file
        retval = list()

        # Values is a nested list of captured information
        for ii in (values, values[0]):
            if not isinstance(ii, list):
                continue
            for row in ii:
                try:
                    # Require the row to be a list
                    assert isinstance(row, list)
                    # Require row to be exactly as long as the header list
                    assert len(row) == len(header)
                    row_dict = {}
                    for idx, value in enumerate(row):
                        row_dict[header[idx]] = value
                    retval.append(row_dict)
                except AssertionError:
                    break
            if len(retval) > 0:
                return retval
            else:
                raise ValueError("Cannot parse the textfsm template")

    def execute(
        self, cmd="", prompt_list=(), timeout=0, template=None, debug=0, consume=True
    ):
        assert isinstance(cmd, str)
        if cmd.strip() == "":
            assert len(cmd.splitlines()) == 0
        else:
            assert len(cmd.splitlines()) == 1
        assert isinstance(timeout, int)
        assert (template is None) or isinstance(template, str)
        assert isinstance(debug, int)

        cmd = cmd.strip()

        if debug > 0:
            logger.debug("Calling execute(cmd='%s', timeout=%s)" % (cmd, timeout))

        if len(prompt_list) > 0:
            self._extend_prompt(prompt_list)

        normal_timeout = self.conn.get_timeout()
        if timeout > 0:
            self.conn.set_timeout(timeout)

        # Handle prompt_list...
        self.set_custom_prompts(prompt_list)

        self.json_log_entry(cmd=cmd, action="execute", timeout=False)

        # Handle sudo command...
        if cmd.strip()[0:4] == "sudo":
            pre_sudo_prompts = self.conn.get_prompt()
            # FIXME I removed re.compile from the sudo prompt. Example prompt:
            #    [sudo] password for mpenning:
            sudo_prompt = re.compile(r"[\r\n].+?:")
            prompts_w_sudo = copy.deepcopy(pre_sudo_prompts)
            prompts_w_sudo.insert(0, sudo_prompt)
            self.conn.set_prompt(prompts_w_sudo)

            # Sending sudo cmd here...
            self.conn.send(cmd + os.linesep)
            prompt_idx, re_match_object = self.conn.expect_prompt()
            # idx==0 is a sudo password prompt...
            if prompt_idx == 0:
                self.conn.set_prompt(self.default_prompt_list)
                self.conn.send(self.password + os.linesep)
                prompt_idx, re_match_object = self.conn.expect_prompt()
                assert isinstance(prompt_idx, int)

            else:
                raise ValueError("Cannot complete 'execute(cmd='%s')" % cmd)
            self.conn.set_prompt(pre_sudo_prompts)

        # Handle non-sudo execute()...
        else:
            try:
                self.conn.execute(cmd)

            except ConnectionResetError as dd:
                error = "SSH session with {0} was reset while running cmd='{1}'".format(
                    self.host, cmd
                )
                raise ConnectionResetError(error)

            except InvalidCommandException:

                if self.allow_invalid_command is False:
                    error = "cmd='%s' is an invalid command" % cmd
                    logger.critical(error)
                    raise InvalidCommandException(error)

        # Reset the prompt list at the end of the command...
        if len(prompt_list) > 0:
            self.reset_prompt()

        # Reset the timeout at the end of the command...
        if self.conn.get_timeout() != normal_timeout:
            self.conn.set_timeout(normal_timeout)

        # save the raw response...
        cmd_output = self.conn.response

        self.json_log_entry(cmd=cmd, action="output", result=cmd_output, timeout=False)
        ## TextFSM
        ## If template is specified, parse the response into a list of dicts...
        if isinstance(template, str):
            return self.tfsm(template, cmd_output)
        else:
            return cmd_output

    def send(self, cmd="", debug=0):
        assert isinstance(cmd, str)
        assert len(cmd.splitlines()) == 1
        assert isinstance(debug, int)

        self.json_log_entry(cmd=cmd, action="send", result=None, timeout=False)

        self.conn.send(cmd)

    def expect(self, prompt_list=(), timeout=0, debug=0):
        """Expect prompts, including those in prompt_list"""
        assert isinstance(debug, int)
        if debug > 0:
            pass

        normal_timeout = self.conn.get_timeout()
        if timeout > 0:
            self.conn.set_timeout(timeout)

        normal_prompts = self.conn.get_prompt()

        # Handle prompt_list...
        self.set_custom_prompts(prompt_list)

        prompt_idx, re_match_object = self.conn.expect_prompt()

        self.conn.set_timeout(normal_timeout)
        # Reset the prompt list at the end of the command...
        if len(prompt_list) > 0:
            self.reset_prompt()

        self.json_log_entry(cmd=cmd, action="expect", result=None, timeout=False)
        return prompt_idx, re_match_object

    def set_timeout(self, timeout=0):
        """Set the command timeout"""
        if isinstance(timeout, int) and timeout > 0:
            return self.conn.set_timeout(timeout)

    def set_custom_prompts(self, prompt_list=()):
        """Wrapper around set_prompt()"""
        return self.set_prompt(prompt_list)

    def reset_prompt(self):
        """Reset all prompts to default"""
        return self.conn.set_prompt(self.default_prompt_list)

    def set_prompt(self, prompt_list=()):
        """Extend the expected prompts with prompts in prompt_list"""
        normal_prompts = self.conn.get_prompt()
        custom_prompts = copy.copy(normal_prompts)
        if (isinstance(prompt_list, tuple) or isinstance(prompt_list, list)) and len(
            prompt_list
        ) > 0:
            prompt_list.reverse()
            for prompt in prompt_list:
                if isinstance(prompt, str):
                    custom_prompts.insert(0, re.compile(prompt))

                elif isinstance(prompt, re.Pattern):
                    custom_prompts.insert(0, prompt)

                else:
                    raise ValueError("Cannot process prompt='%s'" % prompt)
            self.conn.set_prompt(custom_prompts)

        return normal_prompts, custom_prompts

    def close(self, force=True, timeout=0, debug=0):
        self.conn.close(force=force)

    @property
    def response(self):
        return self.conn.response


@logger.catch(default=True, onerror=lambda _: sys.exit(1))
def reset_conn_parameters(conn=None):
    assert isinstance(conn, SSH2)
    conn.set_connect_timeout(DEFAULT_CONNECT_TIMEOUT)
    conn.set_prompt(DEFAULT_PROMPT_LIST)
    conn.set_timeout(DEFAULT_PROMPT_TIMEOUT)


if __name__ == "__main__":
    pass



from getpass import getpass, getuser
from io import StringIO
import time
import copy
import sys
import re
import os

from traits.api import (
    Any,
    CStr,
    Str,
    Bool,
    Int,
    File,
    Undefined,
    ReadOnly,
    Disallow,
    PrefixList,
)
from traits.api import PrefixList, List, Range, Subclass
from traits.api import HasRequiredTraits
from traits.api import Constant

# Base class for server tests...
# - Exscript/tests/Exscript/protocols/ProtocolTest.py
#
# Test suite for the SSH2 client...
# - Exscript/tests/Exscript/protocols/SSH2Test.py
#
# Test suite for the Dummy client...
# - Exscript/tests/Exscript/protocols/DummyTest.py

from Exscript.protocols.exception import ExpectCancelledException
from Exscript.protocols.exception import InvalidCommandException
from Exscript.protocols.exception import TimeoutException  # <-- expect timeout
from Exscript.util.match import any_match
from Exscript import Account, PrivateKey
from Exscript.util.log import log_to
from Exscript.protocols import SSH2
import Exscript

from textfsm import TextFSM

from loguru import logger

"""
Copyright 2021 - David Michael Pennington
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1.  Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

2.  Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

3.  Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from
    this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

"""
TODO:

1. Incorporate tests against the Cisco free router testbeds
described by bigevilbeard, below...

- https://github.com/bigevilbeard/Basic_NetMiko/README.md
  - https://devnetsandbox.cisco.com/RM/Diagram/Index/e83cfd31-ade3-4e15-91d6-3118b867a0dd?diagramType=Topology
  - https://devnetsandbox.cisco.com/RM/Diagram/Index/38ded1f0-16ce-43f2-8df5-43a40ebf752e?diagramType=Topology

2. Fix Exscript SSH with legacy ciphers and kexexchange algorithms -> https://github.com/knipknap/exscript/issues/215
"""

HOST = ""
USERNAME = getuser()
PASSWORD = ""
DEFAULT_LOGIN_TIMEOUT = 0
DEFAULT_PROMPT_LIST = []
DEFAULT_PROMPT_TIMEOUT = 30


def cast_unicode(line, encoding="utf-8"):
    assert (
        isinstance(line, int)
        or isinstance(line, bytes)
        or isinstance(line, str)
        or isinstance(line, float)
    )
    try:
        line = f"{line}"
    except:
        # line = str(bytes(line, encoding='utf-8'), encoding='utf-8')
        if isinstance(line, bytes):
            line = line.decode("utf-8")
        line = str(line, encoding="utf-8")
    return line


@logger.catch
class Shell(HasRequiredTraits):
    host = Str(value="", required=True)
    port = Range(value=22, low=1, high=65524)
    username = Str(value=getuser(), required=False)
    password = Str(required=False)
    private_key_path = File(value=os.path.expanduser("~/.ssh/id_rsa"))
    # FIXME - needs more drivers... -> https://exscript.readthedocs.io/en/latest/Exscript.protocols.drivers.html
    driver = PrefixList(
        value="generic", values=["generic", "shell", "junos", "ios"], required=False
    )
    termtype = PrefixList(
        value="dumb", values=["dumb", "xterm", "vt100"], required=False
    )
    protocol = PrefixList(value='ssh', values=['ssh'])
    stdout = PrefixList(value=None, values=[None, sys.stdout], required=False)
    stderr = PrefixList(value=sys.stderr, values=[None, sys.stderr], required=False)
    banner_timeout = Range(value=20, low=1, high=30, required=False)
    connect_timeout = Range(value=30, low=1, high=30, required=False)
    prompt_timeout = Range(value=10, low=1, high=65535, required=False)
    prompt_list = List(Str, required=False)
    default_prompt_list = List(re.Pattern, required=False)
    account_list = List(Exscript.account.Account, required=False)
    encoding = PrefixList(value="utf-8", values=["latin-1", "utf-8"], required=False)
    conn = Any(required=False)
    debug = Range(value=0, low=0, high=5, required=False)

    def __init__(self, **kwargs):
        HasRequiredTraits.__init__(self, **kwargs)

        assert self.host != ""
        assert len(self.account_list) == 0
        self.append_account()

        #self.conn = SSH2(
        #    driver=self.driver,
        #    stdout=self.stdout,
        #    stderr=self.stderr,
        #    termtype=self.termtype,
        #    banner_timeout=self.banner_timeout,
        #    encoding=self.encoding,
        #    debug=self.debug,
        #)

        self.conn = self.do_ssh_login(username=self.username, password=self.password, login_timeout=30, debug=self.debug)

        #self.login()
        self.default_prompt_list = self.conn.get_prompt()

        # Populate the initial prompt list...

    def __repr__(self):
        return """<Shell: %s>""" % self.host

    def do_ssh_login(self,
        username=None,
        password=None,
        login_timeout=30,
        debug=False,
    ):
        assert isinstance(username, str)
        assert isinstance(debug, int)
        # FIXME - clean up al the globals below... they should not be required

        global HOST
        global USERNAME
        global PASSWORD
        global DEFAULT_LOGIN_TIMEOUT
        global DEFAULT_PROMPT_LIST
        global DEFAULT_PROMPT_TIMEOUT

        if password is None and PASSWORD == "":
            PASSWORD = getpass("Host='%s' password for %s: " % (hostname, username))
            self.password = password
            password = PASSWORD

        elif isinstance(password, str):
            self.password = password
            password = PASSWORD

        private_key=PrivateKey(keytype='rsa').from_file(self.private_key_path)
        account = Account(USERNAME, PASSWORD, key=private_key)

        if self.protocol == "ssh":
            conn = SSH2(driver=self.driver)

            DEFAULT_LOGIN_TIMEOUT = conn.get_connect_timeout()
            DEFAULT_PROMPT_LIST = conn.get_prompt()
            DEFAULT_PROMPT_TIMEOUT = conn.get_timeout()

            conn.set_connect_timeout(login_timeout)
            conn.connect(hostname=self.host, port=self.port)
            conn.login(account)

            # Send a blank line to ensure the connection is alive...
            prompt_idx, prompt_re_match, output = do_command(
                cmd="", conn=conn, prompt_timeout=2
            )
            assert isinstance(prompt_idx, int)
            assert isinstance(prompt_re_match, re.Match)
            response = conn.response
            assert isinstance(response, str)

            conn.set_connect_timeout(DEFAULT_LOGIN_TIMEOUT)

            return conn

        else:
            raise ValueError("FATAL: proto='%s' isn't a valid protocol" % proto)

    def deprecated_login(self):
        self.conn.set_connect_timeout(self.connect_timeout)
        self.conn.connect(self.host, port=self.port)
        assert len(self.account_list) > 0
        self.conn.login(self.account_list[0], app_account=None, flush=True)  # <- FIXME this should use any acct

    def append_account(self):
        # From the exscript docs...
        #     key = PrivateKey.from_file('~/.ssh/id_rsa', 'my_key_password')
        self.username = getuser()
        self.password = getpass("Password for %s: " % self.username)
        acct = Account(self.username, self.password)
        self.account_list.append(acct)

    def _extend_prompt(self, prompt_list=()):
        retval = list()
        for ii in prompt_list:
            if isinstance(ii, str):
                compiled = re.compile(ii)
                retval.append(compiled)

            elif isinstance(ii, re.Pattern):
                retval.append(ii)

            else:
                raise ValueError("Cannot process prompt:'%s'" % ii)

        for ii in self.conn.get_prompt():
            retval.append(ii)

        self.conn.set_prompt(retval)

    def interact(self):
        raise NotImplementedError
        print("Delegating processing to a human...")
        finished = False
        while not finished:
            cmd = input()
            self.execute(cmd.strip(), consume=False)
            time.sleep(0.1)

    def tfsm(self, template=None, input_str=None):
        """Run the textfsm template against input_str"""
        assert isinstance(template, str)
        assert isinstance(input_str, str)
        if os.path.isfile(os.path.expanduser(str(template))):
            # open the textfsm template from disk...
            fh = open(template, "r")
        else:
            # build a fake filehandle around textfsm template string
            fh = StringIO(template)
        fsm = TextFSM(fh)
        header = fsm.header
        values = fsm.ParseText(input_str)
        assert values != [], "Could not match any values with the template."

        ## Pack the extracted values into a list of dicts, using keys from
        ##   the header file
        retval = list()

        # Values is a nested list of captured information
        for ii in (values, values[0]):
            if not isinstance(ii, list):
                continue
            for row in ii:
                try:
                    # Require the row to be a list
                    assert isinstance(row, list)
                    # Require row to be exactly as long as the header list
                    assert len(row) == len(header)
                    row_dict = {}
                    for idx, value in enumerate(row):
                        row_dict[header[idx]] = value
                    retval.append(row_dict)
                except AssertionError:
                    break
            if len(retval) > 0:
                return retval
            else:
                raise ValueError("Cannot parse the textfsm template")

    def execute(self, cmd="", prompt_list=(), timeout=0, template=None, debug=0, consume=True):
        assert isinstance(cmd, str)
        assert len(cmd.splitlines()) == 1
        assert isinstance(timeout, int)
        assert (template is None) or isinstance(template, str)
        assert isinstance(debug, int)

        cmd = cmd.strip()

        if len(prompt_list) > 0:
            self._extend_prompt(prompt_list)

        normal_timeout = self.conn.get_timeout()
        if timeout > 0:
            self.conn.set_timeout(timeout)

        # Handle prompt_list...
        self.set_custom_prompts(prompt_list)

        if cmd.strip()[0:4]=="sudo":
            pre_sudo_prompts = self.conn.get_prompt()
            sudo_prompt = re.compile(r"\s+password\s+[^:]+:")
            prompts_w_sudo = copy.copy(pre_sudo_prompts)
            prompts_w_sudo.insert(0, sudo_prompt)
            self.conn.set_prompt(prompts_w_sudo)
            self.conn.send(cmd+os.linesep)
            prompt_idx, re_match_object = self.conn.expect_prompt()
            # idx==0 is a sudo password prompt...
            if prompt_idx==0:
                self.conn.send(self.password+os.linesep)
                prompt_idx = self.conn.expect(prompts_w_sudo)
                if prompt_idx == 0:
                    raise ValueError("Invalid sudo password")
            elif prompt_idx > 0:
                pass
            else:
                raise ValueError("Cannot complete 'execute(cmd='%s')" % cmd)
            self.conn.set_prompt(pre_sudo_prompts)

        else:
            self.conn.execute(cmd)

        # Reset the prompt list at the end of the command...
        if len(prompt_list) > 0:
            self.reset_prompt()

        # Reset the timeout at the end of the command...
        if self.conn.get_timeout() != normal_timeout:
            self.conn.set_timeout(normal_timeout)

        ## TextFSM
        ## If template is specified, parse the response into a list of dicts...
        if isinstance(template, str):
            return self.tfsm(template, self.conn.response)
        else:
            return self.conn.response

    def send(self, cmd="", debug=0):
        assert isinstance(cmd, str)
        assert len(cmd.splitlines()) == 1
        assert isinstance(debug, int)

        self.conn.send(cmd)

    def expect(self, prompt_list=(), timeout=0, debug=0):
        """Expect prompts, including those in prompt_list"""

        normal_timeout = self.conn.get_timeout()
        if timeout > 0:
            self.conn.set_timeout(timeout)

        normal_prompts = self.conn.get_prompt()

        # Handle prompt_list...
        self.set_custom_prompts(prompt_list)

        prompt_idx, re_match_object = self.conn.expect_prompt()

        self.conn.set_timeout(normal_timeout)
        # Reset the prompt list at the end of the command...
        if len(prompt_list) > 0:
            self.reset_prompt()

        return prompt_idx, re_match_object

    def set_timeout(self, timeout=0):
        """Set the command timeout"""
        if isinstance(timeout, int) and timeout>0:
            return self.conn.set_timeout(timeout)

    def set_custom_prompts(self, prompt_list=()):
        """Wrapper around set_prompt()"""
        return self.set_prompt(prompt_list)

    def reset_prompt(self):
        """Reset all prompts to default"""
        return self.conn.set_prompt(self.default_prompt_list)

    def set_prompt(self, prompt_list=()):
        """Extend the expected prompts with prompts in prompt_list"""
        normal_prompts = self.conn.get_prompt()
        custom_prompts = copy.copy(normal_prompts)
        if (isinstance(prompt_list, tuple) or isinstance(prompt_list, list)) and len(prompt_list) > 0:
            prompt_list.reverse()
            for prompt in prompt_list:
                if isinstance(prompt, str):
                    custom_prompts.insert(0, re.compile(prompt))

                elif isinstance(prompt, re.Pattern):
                    custom_prompts.insert(0, prompt)

                else:
                    raise ValueError("Cannot process prompt='%s'" % prompt)
            self.conn.set_prompt(custom_prompts)

        return normal_prompts, custom_prompts

    def close(self, force=False, timeout=0, debug=0, consume=True):
        self.conn.close(force=force)

    @property
    def response(self):
        return self.conn.response

@logger.catch
def reset_conn_parameters(conn=None):
    assert isinstance(conn, SSH2)
    conn.set_connect_timeout(DEFAULT_LOGIN_TIMEOUT)
    conn.set_prompt(DEFAULT_PROMPT_LIST)
    conn.set_timeout(DEFAULT_PROMPT_TIMEOUT)


@logger.catch
def do_command(cmd=None, conn=None, prompt_list=(), prompt_timeout=30):
    assert isinstance(cmd, str)

    output = None
    match_prompt_idx = None

    if prompt_list == () or prompt_list == []:
        conn.set_prompt(DEFAULT_PROMPT_LIST)

    conn.set_timeout(prompt_timeout)
    try:
        prompt_idx, prompt_re_match = conn.execute(cmd)

    except TimeoutException as ee:
        error = "Host=%s timeout after %s seconds while %s waiting for %s" % (
            "FIXME",
            prompt_timeout,
            cmd,
        )
        raise TimeoutException(error)

    output = conn.response
    assert isinstance(output, str)

    reset_conn_parameters(conn=conn)

    # FIXME - add matching prompt index and regex here...
    return prompt_idx, prompt_re_match, output




#############################################################################
#############################################################################
#############################################################################
#############################################################################
##########################    Backups here      #############################
#############################################################################
#############################################################################
#############################################################################
#############################################################################
#############################################################################

def do_login(
    hostname="localhost",
    driver="generic",
    username=None,
    password=None,
    proto="ssh",
    login_timeout=30,
    debug=False,
):
    assert isinstance(username, str)
    assert isinstance(debug, bool)

    global HOST
    global USERNAME
    global PASSWORD
    global DEFAULT_LOGIN_TIMEOUT
    global DEFAULT_PROMPT_LIST
    global DEFAULT_PROMPT_TIMEOUT

    # Set the global hostname var...
    HOST = hostname

    if password is None and PASSWORD == "":
        PASSWORD = getpass("Host='%s' password for %s: " % (hostname, username))
        password = PASSWORD

    elif isinstance(password, str):
        pass

    account = Account(USERNAME, PASSWORD)

    if proto == "ssh":
        conn = SSH2(driver=driver)

        DEFAULT_LOGIN_TIMEOUT = conn.get_connect_timeout()
        DEFAULT_PROMPT_LIST = conn.get_prompt()
        DEFAULT_PROMPT_TIMEOUT = conn.get_timeout()

        conn.set_connect_timeout(login_timeout)
        conn.connect("localhost")
        conn.login(account)

        # Send a blank line to ensure the connection is alive...
        prompt_idx, prompt_re_match, output = do_command(
            cmd="", conn=conn, prompt_timeout=2
        )
        assert isinstance(prompt_idx, int)
        assert isinstance(prompt_re_match, re.Match)
        response = conn.response
        assert isinstance(response, str)

        conn.set_connect_timeout(DEFAULT_LOGIN_TIMEOUT)

        return conn

    else:
        raise ValueError("FATAL: proto='%s' isn't a valid protocol" % proto)



@logger.catch
def main():

    # login to this system and demo a few commands...
    conn = Shell(username="mpenning", host="localhost")
    output = conn.execute("ls -la | wc -l")

    # conn.set_prompt([r"\]\#", r"\]\$"])
    conn.set_prompt([r"---", r"==="])
    conn.set_timeout(1)
    uname_output = conn.execute("uname -a")

    print("PWD", conn.execute("pwd"))
    # resetting prompts after intentionally adding junk prompt matches above...
    conn.reset_prompt()

    # sudo example...
    conn.set_timeout(2)
    output = conn.execute("sudo ls -la")

    # Parse conn.response for file name and permissions... build a dict with the results
    file_permission_dict = dict()
    for permission_str, filename in any_match(output, r"^\s*(\S+)\s.+?\s+(\S+)\s*$"):
        print("BAR", permission_str, filename)

    conn.send("exit\r")
    conn.close()

if __name__=="__main__":
    main()



from contextlib import closing
from io import StringIO
import unicodedata
import platform
import socket
import uuid
import time
import sys
import re
import os

assert sys.version_info >= (3, 0, 0), "cannon does not support Python 2"

from rich import print as rich_print
from textfsm import TextFSM
import pexpect as px
import transitions

# import snoop

r"""
cannon - Python ssh automation
Copyright (C) 2022      David Michael Pennington
Copyright (C) 2020-2021 David Michael Pennington at Cisco Systems

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

1.  Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

2.  Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

3.  Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from
    this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


class UnexpectedPrompt(Exception):
    """Exception for an Unexpected Prompt"""

    def __init__(self, error=""):
        super(UnexpectedPrompt, self).__init__(error)


class PromptDetectionError(Exception):
    """Exception for an Unexpected Prompt"""

    def __init__(self, error=""):
        super(PromptDetectionError, self).__init__(error)


class InvalidLogFile(Exception):
    """Exception for an invalid log file"""

    def __init__(self, error=""):
        super(InvalidLogFile, self).__init__(error)


# DO NOT escape '$' here
EXPECTED_LAST_PROMPT_CHARS = (":", ">", "#", "$")
BASE_PROMPT_REGEX_LENGTH = 7


class TeeStdoutFile(object):
    """Simple class to send stdout to screen and log_file simultaneously"""

    def __init__(self, log_file="", filemode="w", log_screen=False, encoding="utf-8"):
        self.log_file = os.path.expanduser(log_file)
        self.filemode = filemode
        self.log_screen = log_screen
        self.stdout = sys.stdout
        self.encoding = encoding

        # automatically over-write empty log files
        if os.path.isfile(self.log_file):
            log_file_size = int(os.path.getsize(self.log_file))
            if log_file_size == 0:
                os.remove(self.log_file)
            else:
                raise InvalidLogFile("FATAL: can not overwrite %s" % self.log_file)

        self.fh = open(self.log_file, self.filemode, encoding=self.encoding)

    def __del__(self):
        sys.stdout = self.stdout
        try:
            self.fh.close()
        except AttributeError:
            # We hit this if self.fh was never opened such as existing log_file
            pass

    def write(self, line):
        for ii in line:
            if isinstance(ii, bytes):
                self.fh.write(ii.decode(self.encoding))
                self.stdout.write(ii.decode(self.encoding))
            elif isinstance(ii, str):
                self.fh.write(ii)
                self.stdout.write(ii)

    def flush(self):
        self.fh.flush()

    def close(self):
        self.fh.close()

# Globals used by child.interact()...
#     https://stackoverflow.com/a/43383182/667301
filter_buf = ''
filter_buf_size = 256
let_me_out = False
bash_prompt = re.compile('linux>')

class Account(object):
    def __init__(self, user, passwd="", priv_passwd="", ssh_key=""):
        self.user = user
        self.passwd = passwd
        self.priv_passwd = priv_passwd
        self.ssh_key = os.path.expanduser(ssh_key)  # Path to ssh private key

    def __repr__(self):
        return """<Account user:{} passwd:{} priv_passwd:{} ssh_key: {}>""".format(
            self.user, self.passwd, self.priv_passwd, self.ssh_key
        )


class Shell(transitions.Machine):
    def __init__(
        self,
        host="",
        credentials=(),
        ssh_keepalive=60,
        protocols=({"proto": "ssh", "port": 22}, {"proto": "telnet", "port": 23}),
        mode="",
        auto_priv_mode=None,
        log_screen=False,
        log_file="",
        strip_colors=True,
        debug=False,
        command_timeout=30,
        login_timeout=10,
        relogin_delay=120,
        encoding="utf-8",
        login_attempts=3,
    ):

        STATES = (
            "INIT_SESSION",
            "SELECT_TCP_PROTOCOL",
            "SELECT_LOGIN_CREDENTIALS",
            "SEND_LOGIN_USERNAME",
            "SEND_LOGIN_PASSWORD",
            "CONNECT",
            "LOGIN_SUCCESS_UNPRIV",
            "LOGIN_SUCCESS_PRIV",
            "LOGIN_TIMEOUT",
            "SEND_PRIV_PASSWORD",
            "LOGIN_COMPLETE",
            "INTERACT",
            "CLOSE_SESSION",
        )
        super(Shell, self).__init__(states=STATES, initial="INIT_SESSION")

        self.host = host
        self.credentials = credentials
        self.protocols = protocols
        self.auto_priv_mode = auto_priv_mode
        if auto_priv_mode is not None:
            rich_print(
                "[bold red]WARNING: auto_priv_mode will be deprecated.  Please stop using this option.[/bold red]"
            )
            time.sleep(2.5)
        self.log_screen = log_screen
        self.log_file = os.path.expanduser(log_file)
        self.debug = debug
        self.command_timeout = command_timeout
        self.login_timeout = login_timeout
        self.relogin_delay = relogin_delay
        self.encoding = encoding
        self.login_attempts = login_attempts

        self.child = None  # Pexpect's child object
        self.username = None
        self.password = None
        self.ssh_key = ""
        self.ssh_keepalive = int(ssh_keepalive)
        self.ciphers = ""
        self.key_exchanges = ""
        self.connect_cmd = ""
        self.credentials_iterator = self.iter_credentials()
        self.proto_dict = {}

        self.mode = mode
        self.strip_colors = strip_colors

        self.prompt_hostname = ""
        # Detect a typical linux CLI prompt...
        # Build the template before detecting prompt
        self.base_prompt_regex = self.build_base_prompt_regex()

        self.matching_prompt_regex = ""
        self.matching_prompt_regex_index = -1
        self.matching_string = ""

        #######################################################################
        ## Transitions to SELECT_TCP_PROTOCOL state
        #######################################################################
        self.add_transition(
            trigger="_go_SELECT_TCP_PROTOCOL",
            source="INIT_SESSION",
            dest="SELECT_TCP_PROTOCOL",
            after="after_SELECT_TCP_PROTOCOL_cb",
        )

        #######################################################################
        ## Transitions to SELECT_LOGIN_CREDENTIALS state
        #######################################################################
        self.add_transition(
            trigger="_go_SELECT_LOGIN_CREDENTIALS",
            source="SELECT_TCP_PROTOCOL",
            dest="SELECT_LOGIN_CREDENTIALS",
            after="after_SELECT_LOGIN_CREDENTIALS_cb",
        )

        self.add_transition(
            trigger="_go_SELECT_LOGIN_CREDENTIALS",
            source="SEND_LOGIN_PASSWORD",
            dest="SELECT_LOGIN_CREDENTIALS",
            after="after_SELECT_LOGIN_CREDENTIALS_cb",
        )

        # FIXME - removed on 31 Jan 2021
        # self.add_transition(
        #    trigger="_go_SELECT_LOGIN_CREDENTIALS",
        #    source="LOGIN_COMPLETE",
        #    dest="SELECT_LOGIN_CREDENTIALS",
        #    after="after_SELECT_LOGIN_CREDENTIALS_cb",
        # )

        #######################################################################
        ## Transitions to CONNECT state
        #######################################################################
        self.add_transition(
            trigger="_go_CONNECT",
            source="CONNECT",
            dest="CONNECT",
            after="after_CONNECT_cb",
        )

        self.add_transition(
            trigger="_go_CONNECT",
            source="SELECT_LOGIN_CREDENTIALS",
            dest="CONNECT",
            after="after_CONNECT_cb",
        )

        #######################################################################
        ## Transitions to SEND_LOGIN_USERNAME state
        #######################################################################
        self.add_transition(
            trigger="_go_SEND_LOGIN_USERNAME",
            source="CONNECT",
            dest="SEND_LOGIN_USERNAME",
            after="after_SEND_LOGIN_USERNAME_cb",
        )

        # In case we got to LOGIN_COMPLETE prematurely...
        self.add_transition(
            trigger="_go_SEND_LOGIN_USERNAME",
            source="LOGIN_COMPLETE",
            dest="SEND_LOGIN_USERNAME",
            after="after_SEND_LOGIN_USERNAME_cb",
        )

        #######################################################################
        ## Transitions to SEND_LOGIN_PASSWORD state
        #######################################################################
        self.add_transition(
            trigger="_go_SEND_LOGIN_PASSWORD",
            source="CONNECT",
            dest="SEND_LOGIN_PASSWORD",
            after="after_SEND_LOGIN_PASSWORD_cb",
        )

        self.add_transition(
            trigger="_go_SEND_LOGIN_PASSWORD",
            source="SEND_LOGIN_USERNAME",
            dest="SEND_LOGIN_PASSWORD",
            after="after_SEND_LOGIN_PASSWORD_cb",
        )

        # In case we need to try the same password again
        self.add_transition(
            trigger="_go_SEND_LOGIN_PASSWORD",
            source="SEND_LOGIN_PASSWORD",
            dest="SEND_LOGIN_PASSWORD",
            after="after_SEND_LOGIN_PASSWORD_cb",
        )

        # In case we got to LOGIN_COMPLETE prematurely...
        self.add_transition(
            trigger="_go_SEND_LOGIN_PASSWORD",
            source="LOGIN_COMPLETE",
            dest="SEND_LOGIN_PASSWORD",
            after="after_SEND_LOGIN_PASSWORD_cb",
        )

        #######################################################################
        ## Transitions to LOGIN_SUCCESS_UNPRIV state
        #######################################################################
        self.add_transition(
            trigger="_go_LOGIN_SUCCESS_UNPRIV",
            source="SEND_LOGIN_USERNAME",
            dest="LOGIN_SUCCESS_UNPRIV",
            after="after_LOGIN_SUCCESS_UNPRIV_cb",
        )

        self.add_transition(
            trigger="_go_LOGIN_SUCCESS_UNPRIV",
            source="SEND_LOGIN_PASSWORD",
            dest="LOGIN_SUCCESS_UNPRIV",
            after="after_LOGIN_SUCCESS_UNPRIV_cb",
        )

        self.add_transition(
            trigger="_go_LOGIN_SUCCESS_UNPRIV",
            source="CONNECT",
            dest="LOGIN_SUCCESS_UNPRIV",
            after="after_LOGIN_SUCCESS_UNPRIV_cb",
        )

        #######################################################################
        ## Transitions to SEND_PRIV_PASSWORD state
        #######################################################################
        self.add_transition(
            trigger="_go_LOGIN_SUCCESS_UNPRIV",
            source="LOGIN_SUCCESS_UNPRIV",
            dest="SEND_PRIV_PASSWORD",
            after="after_SEND_PRIV_PASSWORD_cb",
        )

        #######################################################################
        ## Transitions to LOGIN_SUCCESS_PRIV state
        #######################################################################
        self.add_transition(
            trigger="_go_LOGIN_SUCCESS_PRIV",
            source="CONNECT",
            dest="LOGIN_SUCCESS_PRIV",
            after="after_LOGIN_SUCCESS_PRIV_cb",
        )

        self.add_transition(
            trigger="_go_LOGIN_SUCCESS_PRIV",
            source="SEND_LOGIN_PASSWORD",
            dest="LOGIN_SUCCESS_PRIV",
            after="after_LOGIN_SUCCESS_PRIV_cb",
        )

        #######################################################################
        ## Transitions to LOGIN_COMPLETE state
        #######################################################################
        self.add_transition(
            trigger="_go_LOGIN_COMPLETE",
            source="LOGIN_SUCCESS_UNPRIV",
            dest="LOGIN_COMPLETE",
            after="after_LOGIN_COMPLETE_cb",
        )

        self.add_transition(
            trigger="_go_LOGIN_COMPLETE",
            source="LOGIN_SUCCESS_PRIV",
            dest="LOGIN_COMPLETE",
            after="after_LOGIN_COMPLETE_cb",
        )

        self.add_transition(
            trigger="_go_LOGIN_COMPLETE",
            source="LOGIN_COMPLETE",
            dest="LOGIN_COMPLETE",
            after="after_LOGIN_COMPLETE_cb",
        )

        #######################################################################
        ## Transitions to INTERACT state
        #######################################################################
        self.add_transition(
            trigger="_go_INTERACT",
            source="LOGIN_COMPLETE",
            dest="INTERACT",
            after="after_INTERACT_cb",
        )

        #######################################################################
        ## Unconditionally transition to the SELECT_TCP_PROTOCOL state
        #######################################################################
        self._go_SELECT_TCP_PROTOCOL()

    def interact_output_filter(self, s):
        """Use this as an interact() output filter"""
        global proc, bash_prompt, filter_buf, filter_buf_size, let_me_out

        filter_buf += str(s)
        filter_buf = filter_buf[-filter_buf_size:]

        if "LET ME OUT" in filter_buf:
            let_me_out = True

        if bash_prompt.search(filter_buf):
            if let_me_out:
                self.child.sendline('exit')
                self.child.expect(pexpect.EOF)
                self.child.wait()
            else:
                self.child.sendline('python')

        return str(s)

    def build_connect_cmd(self):
        if self.debug:
            rich_print("")
            rich_print("    [bold cyan]in build_connect_cmd()[/bold cyan]")
            rich_print(
                "    [bold blue]build_connect_cmd() is configuring ssh parameters for self.connect_cmd[/bold blue]"
            )
            rich_print(
                "    [bold blue]CURRENT STATE:[/bold blue] '[bold magenta]{}[/bold magenta]'".format(
                    self.state
                )
            )

        if self.ciphers:
            cipher_opt = "-c {}".format(self.ciphers)
        else:
            cipher_opt = ""

        if self.key_exchanges:
            key_exchange_opt = "-o KexAlgorithms={}".format(self.key_exchanges)
        else:
            key_exchange_opt = ""

        # Implement ssh or telnet command... ssh with public key
        if self.proto_dict["proto"] == "ssh" and self.ssh_key != "":
            self.connect_cmd = "ssh {} {} -l {} -p {} -i {} -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -o ServerAliveInterval={} {}".format(
                cipher_opt,
                key_exchange_opt,
                self.username,
                self.proto_dict["port"],
                self.ssh_key,
                self.ssh_keepalive,
                self.host,
            )

        elif self.proto_dict["proto"] == "ssh" and self.ssh_key == "":
            # https://serverfault.com/a/1002182/78702
            self.connect_cmd = "ssh {} {} -l {} -p {} -o PubkeyAuthentication=no -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -o ServerAliveInterval={} {}".format(
                cipher_opt,
                key_exchange_opt,
                self.username,
                self.proto_dict["port"],
                self.ssh_keepalive,
                self.host,
            )

        elif self.proto_dict["proto"] == "telnet":
            self.connect_cmd = "telnet {} {}".format(self.host, self.proto_dict["port"])

        else:
            raise NotImplementedError("")

        if self.debug:
            rich_print("    [bold blue]build_connect_cmd() finished[/bold blue]")
            rich_print("")

    def execute(
        self,
        cmd=None,
        template=None,
        prompts=(),
        timeout=0.0,
        command_timeout=0.0,
        carriage_return=True,
        relax_prompt=False,
    ):
        """Run a command and optionally parse with a TextFSM template

            - `cmd` is the command to execute
            - `template` is a string with the text of a TextFSM template
            - `prompts` is a tuple of prompt regexs to apply to the output of the command
            - `timeout` is how long we should wait for the command prompt to return
            - `carriage_return` indicates whether the command should be followed with a carriage-return.  The values are either True or False (default is True, meaning the CR will be sent after the command).

        execute() returns a list of dicts if `template` is specified; otherwise
        it returns None.
        """
        assert cmd is not None
        assert self.child.isalive()  # Don't issue commands against a dead conn

        if self.debug:
            rich_print("")
            rich_print("    [bold cyan]in execute()[/bold cyan]")
            rich_print(
                "        [bold blue]CURRENT STATE:[/bold blue] '[bold magenta]{}[/bold magenta]'".format(
                    self.state
                )
            )

        if timeout == 0.0 and command_timeout > 0.0:
            timeout = command_timeout

        elif timeout == 0.0 and command_timeout == 0.0:
            timeout = self.command_timeout

        arg_list = (
            "cmd",
            "template",
            "prompts",
            "timeout",
            "command_timeout",
            "carriage_return",
        )
        arg = list()
        if self.debug:
            # build the debugging string...
            for ii in arg_list:
                if ii == "cmd":
                    arg.append("'" + cmd + "'")
                elif ii == "template" and template is not None:
                    arg.append("template=TRUNCATED".format(template))
                elif ii == "template" and template is None:
                    arg.append("template=None")
                elif ii == "prompts":
                    arg.append("prompts={}".format(prompts))
                elif ii == "timeout":
                    arg.append("timeout={}".format(timeout))
                elif ii == "carriage_return":
                    arg.append("carriage_return={}".format(carriage_return))
            logstr = ", ".join(arg)
            rich_print(
                "        [bold blue]execute([/bold blue][bold green]{}[/bold green][bold blue])[/bold blue]".format(
                    logstr
                )
            )

        if self.debug:
            rich_print(
                "        [bold blue]execute() is running with cmd='{}'[/bold blue]".format(
                    cmd
                )
            )
        if carriage_return:
            if self.debug:
                rich_print(
                    "        [bold blue] self.child.sendline('{}')[/bold blue]".format(
                        cmd
                    )
                )
            self.child.sendline(cmd)
        else:
            if self.debug:
                rich_print(
                    "        [bold blue] self.child.send('{}')[/bold blue]".format(cmd)
                )
            self.child.send(cmd)

        # Extend the list of cli_prompts if `prompts` was specified
        cli_prompts = self.build_base_prompt_regex(relax_prompt=relax_prompt)
        if prompts != ():
            assert isinstance(prompts, tuple)
            cli_prompts.extend(prompts)  # Add command-specific prompts here...

        # Look for prompt matches after executing the command
        if self.debug:
            rich_print(
                "        [bold blue]execute() is calling cexpect()[/bold blue]".format(
                    cmd
                )
            )

        index = self.cexpect(cli_prompts, timeout=timeout)

        if self.debug:
            rich_print(
                "        [bold blue]execute() matched index %s: %s in cexpect()[/bold blue]"
                % (index, cli_prompts[index])
            )

        ## If template is specified, parse the response into a list of dicts...
        if template is not None:
            if os.path.isfile(os.path.expanduser(str(template))):
                # open the textfsm template from disk...
                fh = open(template, "r")
            else:
                # build a fake filehandle around textfsm template string
                fh = StringIO(template)
            fsm = TextFSM(fh)
            header = fsm.header
            values = fsm.ParseText(self.response)
            assert values != [], "Could not match any values with the template."

            ## Pack the extracted values into a list of dicts, using keys from
            ##   the header file
            retval = list()

            # Values is a nested list of captured information
            for ii in (values, values[0]):
                if not isinstance(ii, list):
                    continue
                for row in ii:
                    try:
                        # Require the row to be a list
                        assert isinstance(row, list)
                        # Require row to be exactly as long as the header list
                        assert len(row) == len(header)
                        row_dict = {}
                        for idx, value in enumerate(row):
                            row_dict[header[idx]] = value
                        retval.append(row_dict)
                    except AssertionError:
                        break
                if len(retval) > 0:
                    return retval
        else:
            return index

    def csendline(self, text):
        if self.debug:
            rich_print("")
            rich_print("    [bold cyan]in csendline()[/bold cyan]")
        assert self.child.isalive()
        if self.debug:
            rich_print("")
            rich_print(
                "            [bold blue]CURRENT STATE:[/bold blue] '[bold magenta]{}[/bold magenta]'".format(
                    self.state
                )
            )
            rich_print(
                "            [bold blue]csendline([/bold blue][bold yellow]'{}'[/bold yellow][bold blue])[/bold blue]".format(
                    text
                )
            )

        if self.debug:
            rich_print(
                "            [bold blue]csendline() calling self.child.sendline()[/bold blue]"
            )
            rich_print(
                "            [bold blue]exiting csendline()[/bold blue]"
            )
        # WARNING: use self.child.sendline(); do not use self.csendline() here
        self.child.sendline(text)

    def cexpect(self, pattern_list, timeout=-1):
        if self.debug:
            rich_print("")
            rich_print("    [bold cyan]in cexpect()[/bold cyan]")
            rich_print(
                "        [bold blue]CURRENT STATE:[/bold blue] '[bold magenta]{}[/bold magenta]'".format(
                    self.state
                )
            )
            rich_print(
                "        [bold blue]cexpect([/bold blue][bold green]pattern_list, timeout={}[/bold green][bold blue]) was called[/bold blue]".format(
                    timeout
                )
            )
            # Expand all pattern_list terms...
            rich_print("        [bold blue]pattern_list = {}[/bold blue]".format("[["))
            for idx, term in enumerate(pattern_list):
                rich_print(
                    "            [bold blue]{} {},[/bold blue]".format(idx, repr(term))
                )
            rich_print("        [bold blue]{}[/bold blue]".format(str("]]")))

        # FIXME - this could break cexpect if ssh connection is still in progress
        # assert self.child.isalive()

        now = time.time()

        try:
            if self.debug:
                rich_print(
                    "        [bold blue]cexpect() calling self.child.expect(pattern_list, timeout={})[/bold blue]".format(
                        timeout
                    )
                )
            match_index = self.child.expect(pattern_list, timeout=timeout)
            self.matching_prompt_regex_index = match_index
            self.matching_prompt_regex = pattern_list[match_index]

            if self.debug:
                delta_secs = round(time.time() - now, 4)
                rich_print("")
                rich_print(
                    "        [bold blue]^^ cexpect() matched regex at match_index {}={}[/bold blue]".format(
                        match_index, repr(pattern_list[match_index])
                    )
                )
                # FIXME why can't I use quotes around self.matching_prompt in rich_print()
                # rich_print(
                #    "      [bold blue]^^ cexpect() matching_prompt={}[/bold blue]".format(
                #        repr(self.matching_prompt)
                #    )
                # )
                rich_print(
                    "        [bold blue]^^ cexpect() match time: {} seconds[/bold blue]".format(
                        delta_secs
                    )
                )
                rich_print("")
                rich_print(
                    "    [bold blue]cexpect() debugs from str(self.child):[/bold blue]"
                )
                for line in str(self.child).splitlines():
                    rich_print("        [bold yellow]{0}[/bold yellow]".format(line))

        except px.exceptions.EOF:
            if self.debug:
                rich_print(
                    "        [bold red]cexpect() EOF exception while waiting for pattern_list[/bold red]"
                )
            match_index = None

        except px.exceptions.TIMEOUT:
            if self.debug:
                rich_print(
                    "        [bold red]cexpect() TIMEOUT exception while waiting for pattern_list[/bold red]"
                )
            match_index = None

        if (match_index is not None) and match_index >= 0:
            try:
                if self.debug:
                    rich_print(
                        "        [bold blue]cexpect() len(pattern_list[match_index])={}".format(
                            len(pattern_list[match_index])
                        )
                    )
                assert not ("UUID~" in pattern_list[match_index][0:6])

            except AssertionError as ee:
                rich_print(
                    "[bold red]cexpect() matched prompt index {}: {}.  {}[/bold red]".format(
                        match_index, pattern_list[match_index], str(ee)
                    )
                )

            return match_index

        else:
            # match_index hit an error... try again...
            self.child.sendline("")
            match_index = self.child.expect(pattern_list, timeout=timeout)

            assert match_index >= 0
            # match_index < 0
            if self.debug:
                rich_print("")
                rich_print(
                    "        [bold blue]cexpect() detected a prompt and is returning '{}' to the calling function[/bold blue]".format(
                        match_index
                    )
                )
            return match_index

    def interact(self):
        self._go_INTERACT()

    def modify_ssh_parameters(self):
        """This method should be called after calling self.cexpect() during ssh spawn.  It returns True or False"""

        if self.debug:
            rich_print("")
            rich_print("    [bold cyan]in modify_ssh_parameters()[/bold cyan]")
            rich_print(
                "    [bold blue]CURRENT STATE:[/bold blue] '[bold magenta]{}[/bold magenta]'".format(
                    self.state
                )
            )

        #  no matching ssh cipher found. Their offer:
        mm = re.search(r"no\s+matching\s+cipher.+?offer:\s+(\S.+)$", self.child.after)
        if mm is not None:
            candidate_cipher_list = mm.group(1).strip()
            self.ciphers = candidate_cipher_list
            if self.debug:
                rich_print(
                    "    [bold blue]modify_ssh_parameters() set self.ciphers: '{}'[/bold blue]".format(
                        self.ciphers
                    )
                )
            # We didn't have a matching prompt character...
            if self.debug:
                rich_print("")
                rich_print(
                    "    [bold blue]modify_ssh_parameters() returning True[/bold blue]"
                )
            return True

        #  no matching key exchange method found. Their offer: diffie-hellman-group14-sha1
        mm = re.search(
            r"no\s+matching\s+key\s+exchange.+?offer:\s+(\S.+)$", self.child.after
        )
        if mm is not None:
            self.key_exchanges = mm.group(1).strip()
            if self.debug:
                rich_print(
                    "    [bold blue]modify_ssh_parameters() set self.key_exchanges.[/bold blue]"
                )
                rich_print(
                    "        valid [bold yellow]key_exchanges={}[/bold yellow]".format(
                        self.key_exchanges
                    )
                )
            # We didn't have a matching prompt character...
            if self.debug:
                rich_print("")
                rich_print(
                    "    [bold blue]modify_ssh_parameters() returning True[/bold blue]"
                )
            return True

        if self.debug:
            rich_print("")
            rich_print(
                "    [bold blue]modify_ssh_parameters() returning False[/bold blue]"
            )
        return False

    @property
    def matching_prompt(self):
        """Get the matching prompt character; return the prompt character. Return None is session is not alive"""
        if self.debug:
            rich_print("    [bold cyan]in matching_prompt()[/bold cyan]")

        # After finds the string which matched...
        after = self.child.after

        self.matching_string = after

        if self.debug:
            rich_print(
                "    [bold blue]matching_prompt() is parsing: '{}'[/bold blue]".format(
                    after
                )
            )

        # Check for a valid CLI prompt character...
        candidate_prompt = None
        if self.child.isalive():
            if self.debug:
                rich_print(
                    "    [bold blue]matching_prompt() is iterating over lines in after.splitlines()[/bold blue]"
                )

            # Process ssh session output line by line...
            for line in after.splitlines():
                if self.debug:
                    rich_print(
                        "        [bold blue]checking line='{}'[/bold blue]".format(line)
                    )
                line = line.strip()
                if len(line) > 0:

                    candidate_prompt = line[-1]
                    prompt_hostname = line.strip()[0:-1]

                    if self.debug:
                        rich_print("    [bold blue]line='{}'[/bold blue]".format(line))
                        rich_print(
                            "    [bold yellow]    candidate_prompt='{}'[/bold yellow]".format(
                                candidate_prompt
                            )
                        )
                        rich_print(
                            "    [bold yellow]    prompt_hostname='{}'[/bold yellow]".format(
                                prompt_hostname
                            )
                        )

            if candidate_prompt in EXPECTED_LAST_PROMPT_CHARS:
                if self.debug:
                    rich_print(
                        "    [bold blue]matching_prompt() is returning candidate_prompt='{}'[/bold blue]".format(
                            candidate_prompt
                        )
                    )
                return candidate_prompt

        else:
            if self.debug:
                rich_print(
                    "    [bold blue]matching_prompt() is returning None='{}'[/bold blue]".format(
                        line
                    )
                )
            candidate_prompt = None
            return candidate_prompt

    @property
    def response(self):
        before = self.child.before
        if self.strip_text_colors:
            return self.strip_text_colors(before)
        else:
            return before

    def exit(self):
        if self.debug:
            rich_print(
                "    [bold blue]exit() is calling self.child.close()[/bold blue]".format(
                    line
                )
            )
        self.child.close()

    def quit(self):
        if self.debug:
            rich_print(
                "    [bold blue]quit() is calling self.exit()[/bold blue]".format(line)
            )
        self.exit()

    def iter_protocols(self):
        if self.debug:
            rich_print("")
            rich_print("    [bold cyan]in iter_protocols()[/bold cyan]")
        for proto_dict in self.protocols:
            yield proto_dict

    def iter_credentials(self):
        if self.debug:
            rich_print("")
            rich_print("    [bold cyan]in iter_credentials()[/bold cyan]")
        for cred in self.credentials:
            if self.debug:
                rich_print("    [bold blue]iter_credentials() is yielding:[/bold blue]")
                rich_print("        [bold yellow]{}[/bold yellow]".format(repr(cred)))
            yield cred

    def sync_prompt(self, require_detect_prompt=True):
        """Catch up with any queued prompts, we know to exit if we get a px.exceptions.TIMEOUT error"""
        if self.debug:
            rich_print("")
            rich_print("    [bold cyan]in sync_prompt()[/bold cyan]")
            rich_print(
                "    [bold blue]CURRENT STATE:[/bold blue] '[bold magenta]{}[/bold magenta]'".format(
                    self.state
                )
            )

        if self.debug:
            rich_print(
                "[bold blue]sync_prompt(require_detect_prompt={}) was called[/bold blue]".format(
                    require_detect_prompt
                )
            )

        # self.detect_prompt() *should* come before self.sync_prompt()
        if (self.prompt_hostname == "") and (require_detect_prompt is True):
            if self.debug:
                rich_print(
                    "[bold blue]detect_prompt() has not assigned a hostname to self.prompt_hostname[/bold blue]".format(
                        self.prompt_hostname
                    )
                )
                rich_print(
                    "[bold red]self.prompt_hostname='{}'.[/bold red]".format(
                        self.prompt_hostname
                    )
                )
            raise PromptDetectionError(
                "Please call detect_prompt() before sync_prompt()"
            )

        # WARNING: use self.child.sendline(); do not use self.csendline() here
        self.child.sendline("")

        finished = False
        while not finished:
            # Use a very short timeout here...
            # WARNING self.child.expect() is required; do not use self.cexpect()
            index = -2
            try:
                index = self.child.expect(self.base_prompt_regex, timeout=2.0)
                if self.debug:
                    rich_print(
                        "    [bold blue]sync_prompt() index={}[/bold blue]".format(
                            index
                        )
                    )

            except px.exceptions.TIMEOUT:
                # We got an EOF or TIMEOUT error...
                self.login_attempts = 0
                finished = True

            except px.exceptions.EOF:
                # We got an EOF or TIMEOUT error...
                self.login_attempts = 0
                finished = True

            except Exception as ee:
                print("ERROR: %s" % ee)

            if index == -2:
                finished = True

            elif index == -1:
                # We got an EOF or TIMEOUT error...
                self.login_attempts = 0
                finished = True

            elif index == 0:
                if self.debug:
                    rich_print(
                        "[bold red]sync_prompt() - found an ssh cipher error[/bold red]".format(
                            index
                        )
                    )

            elif index == 1:
                if self.debug:
                    rich_print(
                        "[bold red]sync_prompt() - found an ssh key exchange error[/bold red]".format(
                            index
                        )
                    )

            elif index == 2 or index == 3:
                if self.debug:
                    rich_print(
                        "[bold red]sync_prompt() - Unexpected index='{}'[/bold red]".format(
                            index
                        )
                    )

            elif index == 4:
                # We should only get to this prompt if auto_priv_mode is
                #     False
                self.login_attempts = 0
                finished = True

            elif index == 5:
                # We should only get to this prompt if auto_priv_mode is
                #     False
                self.login_attempts = 0
                finished = True

            elif index == 6:
                # We don't need to attempt any more logins if we have
                #     a priv prompt
                self.login_attempts = 0
                finished = True

            else:
                raise NotImplementedError()

    def close(self):
        if self.debug:
            rich_print("")
            rich_print("    [bold cyan]in close()[/bold cyan]")
        self.child.close()
        return (self.child.exitstatus, self.child.signalstatus)

    def detect_prompt(self):
        """detect_prompt() checks for premature entry into LOGIN_COMPLETE and also looks for a prompt string"""
        # Detect the prompt as best-possible...
        if self.debug:
            rich_print("")
            rich_print("    [bold cyan]in detect_prompt(mode='{}')[/bold cyan]".format(self.mode))
            rich_print(
                "    [bold blue]CURRENT STATE:[/bold blue] '[bold magenta]{}[/bold magenta]'".format(
                    self.state
                )
            )

        if self.mode == "linux":
            if self.debug:
                rich_print(
                    "    [bold blue]detect_prompt() is calling change_linux_prompt()[/bold blue]"
                )
            # do NOT set detect_prompt=True... that starts an infinite loop...
            self.change_linux_prompt(detect_prompt=False)

            if self.debug:
                rich_print(
                    "    [bold blue]detect_prompt() starting a while loop to catch up with missed expect prompts[/bold blue]"
                )

            # Catch up on missed prompts...
            finished = False
            while not finished:
                try:
                    # FIXME: we need to remove the hard-coded prompt
                    #
                    #     only expect the linux prompt is safe because
                    #     we got self.mode=="linux"
                    self.child.expect([r"linux>"], timeout=1)
                except:
                    rich_print(
                        "        [bold blue]detect_prompt() timed out while expecting prompts with self.child.expect()[/bold blue]"
                    )
                    finished = True

            self.hostname = "linux"
            self.build_base_prompt_regex()  # Adjust the prompt regex after detection
            return self.prompt_hostname

        assert self.mode!="linux"

        if self.debug:
            rich_print(
                "    [bold blue]detect_prompt() using prompt detection heuristics[/bold blue]"
            )
            rich_print(
                "    [bold blue]detect_prompt() is trying to find detailed prompt information.  Sending a blank line to {}[/bold blue]".format(
                    self.host
                )
            )

        self.csendline("")

        # Double check that this isn't a pre-login banner...
        loop_counter = 0
        finished = False
        if self.debug:
            rich_print("    [bold blue]detect_prompt() starting while loop[/bold blue]")
        while not finished:
            # Use a very short timeout here...
            # WARNING use self.child.expect()... do not use self.cexpect()
            if self.debug:
                rich_print(
                    "        [bold blue]in while loop (finished=False)[/bold blue]"
                )
            try:
                loop_counter += 1  # Keep track of how many times we loop through input

                if self.debug:
                    rich_print(
                        "        [bold blue]while loop_counter={}[/bold blue]".format(
                            loop_counter
                        )
                    )
                    rich_print("        [bold blue]calling self.child.expect()[/bold blue]")

                index = self.child.expect(self.base_prompt_regex, timeout=1)

            except px.exceptions.TIMEOUT:
                assert self.child.isalive()
                if self.debug:
                    rich_print("")
                    rich_print(
                            "[bold blue]        detect_prompt() - self.child.expect() timed-out: pexpect.TIMEOUT[/bold blue]"
                    )
                    rich_print(
                            "[bold blue]        detect_prompt() - index=-1.  Exiting while loop[/bold blue]"
                    )
                index = -1
                finished = True

            except px.exceptions.EOF:
                if self.debug:
                    rich_print(
                            "[bold blue]        detect_prompt() - self.child.expect() found pexpect.EOF[/bold blue]"
                    )
                    rich_print(
                            "[bold blue]        detect_prompt() - index=-1.  Exiting while loop[/bold blue]"
                    )
                index = -1
                finished = True

            if index == -1:
                assert self.child.isalive()
                if self.debug:
                    rich_print("")
                    rich_print(
                        "[bold blue]        detect_prompt() index={}[/bold blue]".format(
                            index
                        )
                    )

            elif index == 0 or index == 1:
                # This is an ssh-cipher or ssh key-exchange problem....
                #    fail for now, handle this later if important...
                if self.debug:
                    rich_print("")
                    rich_print(
                        "[bold blue]        detect_prompt() index={}[/bold blue]".format(
                            index
                        )
                    )
                raise NotImplementedError()

            elif index == 2:
                assert self.child.isalive()
                # We probably want to detect_prompt() based on some banner
                # This should be a username prompt...
                if self.debug:
                    rich_print("")
                    rich_print(
                        "[bold blue]        detect_prompt() index={}[/bold blue]".format(
                            index
                        )
                    )
                    rich_print(
                        "        [bold blue]detect_prompt() found a premature entry into detect_prompt().  Redirecting to state SEND_LOGIN_USERNAME[/bold blue]"
                    )
                self._go_SEND_LOGIN_USERNAME()

            elif index == 3:
                assert self.child.isalive()
                # We probably went to detect_prompt() based on some banner
                # This should be a password prompt...
                if self.debug:
                    rich_print("")
                    rich_print(
                        "[bold blue]        detect_prompt() index={}[/bold blue]".format(
                            index
                        )
                    )
                    rich_print(
                        "        [bold blue]detect_prompt() found a premature entry into detect_prompt().  Redirecting to state SEND_LOGIN_PASSWORD[/bold blue]"
                    )

                self._go_SEND_LOGIN_PASSWORD()

            elif index > 3:
                # We got a non-username, non-password prompt... exit this loop!
                # Start prompt detection heuristics...
                if self.debug:
                    # We got a non-username, non-password prompt... exit this loop!
                    rich_print("")
                    rich_print(
                        "        [bold blue]detect_prompt() index={}[/bold blue]".format(
                            index
                        )
                    )
                    rich_print(
                        "        [bold blue]detect_prompt() found a non-username, non-password prompt... exit the while loop.[/bold blue]"
                    )
                    rich_print(
                        "        [bold blue]detect_prompt() set finished=True[/bold blue]"
                    )
                finished = True

            else:
                raise ValueError

            if self.debug:
                rich_print("")
                rich_print(
                    "        [bold blue]detect_prompt() loop_counter={}".format(
                        loop_counter
                    )
                )

        before_retval = list()
        before = self.child.before
        if self.strip_colors:
            before_retval = self.strip_control_chars(before).splitlines()

            # FIXME - I think we can delete this...
            for line in self.strip_text_colors(before).splitlines():
                # https://stackoverflow.com/a/19016117/667301
                no_cntl_char_line = "".join(
                    ch for ch in line if unicodedata.category(ch)[0] != "C"
                )
                before_retval.append(no_cntl_char_line)
        else:
            # I think it's safe to remove these because self.strip_colors==False
            #before = self.strip_control_chars(before)
            #before_retval = self.strip_control_chars(before).splitlines()

            # FIXME - I think we can delete this...
            for line in before.splitlines():
                # https://stackoverflow.com/a/19016117/667301
                no_cntl_char_line = "".join(
                    ch for ch in line if unicodedata.category(ch)[0] != "C"
                )
                before_retval.append(no_cntl_char_line)
        before_stripped = "\r\n".join(before_retval)

        after_retval = list()
        after = self.child.after
        if self.debug:
            rich_print("    [bold blue]detect_prompt() repr(after): {}".format(repr(after)))
            rich_print("        [bold blue]writing output to variable: after_retval")

        if self.strip_colors:
            if self.debug:
                rich_print("        [bold blue]detect_prompt() is stripping colors in variable: after")
                rich_print(
                        "        [bold blue]detect_prompt() after values: '{}'[/bold blue]".format(
                        after
                    )
                )

            for line in self.strip_text_colors(after).splitlines():
                # https://stackoverflow.com/a/19016117/667301
                no_cntl_char_line = "".join(
                    ch for ch in line if unicodedata.category(ch)[0] != "C"
                )
                after_retval.append(no_cntl_char_line)
        else:
            if self.debug:
                rich_print("        [bold blue]detect_prompt() will not strip colors in variable: after")
            for line in after.splitlines():
                # https://stackoverflow.com/a/19016117/667301
                no_cntl_char_line = "".join(
                    ch for ch in line if unicodedata.category(ch)[0] != "C"
                )
                after_retval += no_cntl_char_line
        after_stripped = "\r\n".join(after_retval)

        if self.debug:
            rich_print("")
            rich_print("    [bold blue]detect_prompt(): len(after_stripped)={} lines[/bold blue]".format(len(after_stripped)))
            rich_print(
                "\n    [bold blue]detect_prompt() finished the while loop took {} iterations.  Detected prompt index={}[/bold blue]".format(
                    loop_counter, index
                )
            )

        ## Example of prompt detection on route-views.oregon-ix.org...
        # hostname = self.child.before.strip()  # detect hostname    = route-views
        assert isinstance(after, str)
        if self.debug:
            rich_print("")
            rich_print(
                "        [bold blue]detect_prompt() sent an empty command and saw this in self.child.after: '{}'".format(
                    after_stripped
                )
            )

        assert len(after.splitlines()) > 0

        # Set the hostname
        for line in after_retval:
            if line.strip() == "":
                continue
            else:
                self.prompt_hostname = line.strip()[:-1]
                break

        assert self.prompt_hostname != ""

        if self.debug:
            rich_print(
                "        [bold blue]detect_prompt() searched for the prompt in this output '{}'".format(
                    after_stripped
                )
            )
            rich_print(
                "        [bold yellow]detect_prompt() set prompt_hostname='{}'[/bold yellow]".format(
                    self.prompt_hostname
                )
            )

        assert self.prompt_hostname != ""

        self.build_base_prompt_regex()  # Adjust the prompt regex after detection

        return self.prompt_hostname

    def build_base_prompt_regex(self, relax_prompt=False):
        """Assign self.base_prompt_regex with the latest prompt info"""

        if self.debug:
            rich_print("")
            rich_print("    [bold cyan]in build_base_prompt_regex()[/bold cyan]")

        ####### WARNING #######################################################
        #   Getting this cisco unpriv prompt to work was extremely hard...
        #   there is a bug somewhere that ignores the trailing ">" and
        #   instead matches a space after the hostname.  I forced the
        #   regex to match properly by stripping off the last character of
        #   the hostname and used a negated special character class: [^\n]+?
        #######################################################################
        if relax_prompt is False:
            cisco_unpriv_prompt_str = r"[\r\n]+{0}[^\n]+?{1}".format(
                re.escape(self.prompt_hostname[:-1]), r">"
            )
            linux_prompt_str = r"[\r\n]+{0}[^\n]+?{1}".format(
                re.escape(self.prompt_hostname[:-1]), re.escape("$")
            )
            cisco_priv_prompt_str = r"[\r\n]+{0}[^\n]+?{1}".format(
                re.escape(self.prompt_hostname[:-1]), re.escape("#")
            )

        else:
            cisco_unpriv_prompt_str = r"[\r\n]+[^\n]+?{0}".format(
                r">"
            )
            linux_prompt_str = r"[\r\n]+[^\n]+?{0}".format(
                re.escape("$")
            )
            cisco_priv_prompt_str = r"[\r\n]+[^\n]+?{0}".format(
                re.escape("#")
            )

        self.base_prompt_regex = [
            # Unable to negotiate with 172.16.1.3 port 22: no matching cipher found. Their offer: aes128-cbc,3des-cbc,aes192-cbc,aes256-cbc
            r"no\s+matching\s+cipher\s+found.*$",
            # no matching key exchange method found. Their offer: diffie-hellman-group14-sha1
            r"no\s+matching\s+key\s+exchange\s+method\s+found\..*$",
            r"sername:",
            r"[Pp]assword[^\r\n]{0,20}?:",
            # Cisco unpriv prompt...
            cisco_unpriv_prompt_str,
            # linux prompt
            linux_prompt_str,
            # Cisco priv or linux root prompt...
            cisco_priv_prompt_str,
        ]

        # Remove some prompt matches after successful login
        if self.state in set(
            {
                "LOGIN_SUCCESS_UNPRIV",
                "LOGIN_SUCCESS_PRIV",
                "LOGIN_TIMEOUT",
                "LOGIN_COMPLETE",
            }
        ):

            for index in [0, 1, 2, 3]:
                self.base_prompt_regex[index] = "UUID~" + str(uuid.uuid4())

        if self.debug:
            # Expand all base_prompt_regex terms...
            rich_print(
                "        [bold blue]self.base_prompt_regex = {}[/bold blue]".format(
                    "[["
                )
            )
            for idx, term in enumerate(self.base_prompt_regex):
                rich_print(
                    "            [bold blue]{} {},[/bold blue]".format(idx, repr(term))
                )
            rich_print("        [bold blue]{}[/bold blue]".format(str("]]")))

        assert len(self.base_prompt_regex) == BASE_PROMPT_REGEX_LENGTH
        if self.debug:
            rich_print("    [bold blue]exiting build_base_prompt_regex()[/bold blue]")
        return self.base_prompt_regex

    def change_linux_prompt(self, detect_prompt=True):
        if self.debug:
            rich_print("")
            rich_print("    [bold cyan]in change_linux_prompt()[/bold cyan]")

        self.prompt_hostname = "linux"
        self.base_prompt_regex = self.build_base_prompt_regex()
        if self.debug:
            rich_print("        [bold blue]change_linux_prompt() is running 'precmd_functions=()'[/bold blue]")
        self.child.sendline("precmd_functions=()")
        if self.debug:
            rich_print("        [bold blue]change_linux_prompt() is running 'export PS1={}>'[/bold blue]".format(self.prompt_hostname))
        self.child.sendline("export PS1='{}>'".format(self.prompt_hostname))

        if detect_prompt is True:
            if self.debug:
                rich_print("")
                rich_print(
                    "    [bold blue]change_linux_prompt() is running detect_prompt()[/bold blue]"
                )
            self.detect_prompt()

    def strip_control_chars(self, text_input=""):
        if self.debug:
            rich_print("    [bold cyan]in strip_control_chars(text_input='{}')[/bold cyan]".format(text_input))

        retval = list()
        for line in text_input.splitlines():
            # https://stackoverflow.com/a/19016117/667301
            no_cntl_char_line = "".join(
                ch for ch in line if unicodedata.category(ch)[0] != "C"
            )
            retval.append(no_cntl_char_line)
        return os.linesep.join(retval)

    def strip_text_colors(self, ascii_text=""):
        """This function only works with string inputs, byte inputs fail"""

        if self.debug:
            if isinstance(ascii_text, str) or isinstance(ascii_text, bytes):
                sample = ascii_text[0:10]
                rich_print("")
                rich_print("    [bold cyan]in strip_text_colors(ascii_text='{}...')[/bold cyan]".format(sample))
                rich_print("        [bold blue]ascii_text is type: {}[/bold blue]".format(type(ascii_text)))
        # https://stackoverflow.com/a/14693789/667301...
        assert not isinstance(ascii_text, bytes)
        assert isinstance(ascii_text, str)
        ansi_escape_8bit = re.compile(
            br"(?:\x1B[@-Z\\-_]|[\x80-\x9A\x9C-\x9F]|(?:\x1B\[|\x9B)[0-?]*[ -/]*[@-~])"
        )
        result = ansi_escape_8bit.sub(b"", bytes(ascii_text, self.encoding))
        if self.debug:
            rich_print(
                "        [bold blue]strip_text_colors() returning {} characters[/bold blue]".format(
                    len(result)
                )
            )
        return result.decode(self.encoding)

    def after_SELECT_TCP_PROTOCOL_cb(self):
        """Attempt to make a raw TCP connection to the protocol's TCP port"""

        if self.debug:
            rich_print("")
            rich_print(
                "[bold blue]Entering state: [/bold blue][bold magenta]{}[/bold magenta]".format(
                    self.state
                )
            )

        finished = False
        for ii in range(0, 3):

            if finished is True:
                continue

            if self.debug:
                rich_print("    [bold blue]var ii={}[/bold blue]".format(str(ii)))
            for proto_dict in self.iter_protocols():
                self.proto_dict = proto_dict
                proto_name, proto_port = proto_dict.get("proto"), proto_dict.get("port")
                if self.debug:
                    rich_print(
                        "    [bold blue]Trying TCP socket to {} port {}[/bold blue]".format(
                            self.host, proto_port
                        )
                    )
                with closing(
                    socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                ) as proto_sock:
                    proto_sock.settimeout(3)
                    try:
                        if proto_sock.connect_ex((self.host, proto_port)) == 0:
                            ## We completed a tcp connection to proto_port...
                            ##   now, get the credentials
                            if self.debug:
                                rich_print(
                                    "    [bold blue]SUCCESS: port {}[/bold blue]".format(
                                        proto_port
                                    )
                                )
                            finished = True
                            break
                    except socket.gaierror:
                        raise Exception("'{}' is an unknown hostname".format(self.host))

                    except Exception as ee:
                        raise Exception(ee)

                if ((ii == 0) or (ii == 1)) and (finished is False):

                    time.sleep(
                        self.relogin_delay
                    )  # Give the host time to recover from auto-login

        if finished is False:
            raise Exception("Cannot connect to host: '{}'".format(self.host))
        else:
            if self.debug:
                rich_print(
                    "    [bold blue]after_SELECT_TCP_PROTOCOL_cb() is getting credentials from after_SELECT_LOGIN_CREDENTIALS_cb()[/bold blue]"
                )
            self._go_SELECT_LOGIN_CREDENTIALS()  # Get the proper credentials

    def after_SELECT_LOGIN_CREDENTIALS_cb(self):

        if self.debug:
            rich_print("")
            rich_print(
                "[bold blue]Entering state: [/bold blue][bold magenta]{}[/bold magenta]".format(
                    self.state
                )
            )

        if self.debug:
            rich_print(
                "    [bold blue]after_SELECT_LOGIN_CREDENTIALS_cb() is using a while loop to iterate through attempted credentials[/bold blue]"
            )
        while self.login_attempts > 0:
            if self.debug:
                rich_print(
                    "    [bold blue]after_SELECT_LOGIN_CREDENTIALS_cb() login_attempts={}[/bold blue]".format(
                        self.login_attempts
                    )
                )
            cred = next(self.credentials_iterator)
            if self.debug:
                rich_print(
                    "        [bold blue]after_SELECT_LOGIN_CREDENTIALS_cb() self.username={}[/bold blue]".format(
                        cred.user
                    )
                )
                rich_print(
                    "        [bold blue]after_SELECT_LOGIN_CREDENTIALS_cb() self.password={}[/bold blue]".format(
                        cred.passwd
                    )
                )
                rich_print(
                    "        [bold blue]after_SELECT_LOGIN_CREDENTIALS_cb() self.ssh_key={}[/bold blue]".format(
                        cred.ssh_key
                    )
                )
            self.username = cred.user
            self.password = cred.passwd
            self.ssh_key = cred.ssh_key

            if self.child is not None:
                if self.debug:
                    rich_print(
                        "    [bold yellow]after_SELECT_LOGIN_CREDENTIALS_cb() is closing the current connection[/bold yellow]".format(
                            cred.passwd
                        )
                    )
                self.close()  # Make way for a fresh child instance

            if self.debug:
                rich_print(
                    "    [bold blue]after_SELECT_LOGIN_CREDENTIALS_cb() is calling after_CONNECT_cb()[/bold blue]".format(
                        cred.passwd
                    )
                )
            self._go_CONNECT()

            self.login_attempts -= 1

    def after_CONNECT_cb(self):
        """Build an ssh / telnet session to self.host"""

        if self.debug:
            rich_print("")
            rich_print(
                "[bold blue]Entering state: [/bold blue][bold magenta]{}[/bold magenta]".format(
                    self.state
                )
            )

        if self.child is not None:
            if self.debug:
                rich_print(
                    "[bold yellow]after_CONNECT_cb() is closing the existing connection."
                )
            self.close()

        # run the ssh or telnet command
        ready_for_login = False
        while ready_for_login is False:

            try:
                self.build_connect_cmd()

                # Very strange things happen unless this is here...
                #      Transition from LOGIN_COMPLETE to spawning ssh????
                if self.state == "LOGIN_COMPLETE":
                    if self.debug:
                        rich_print(
                            "[bold magenta]Bypassing strange detour to after_CONNECT_cb()[/bold magenta]".format(
                                self.connect_cmd
                            )
                        )
                    ready_for_login = True
                    return None

                if self.debug:
                    rich_print(
                        "    [bold blue]after_CONNECT_cb() is calling pexpect.spawn()[/bold blue]"
                    )
                    rich_print(
                        "    [bold blue]pexpect.spawn([/bold blue]'{}'[bold blue])[/bold blue]".format(
                            self.connect_cmd
                        )
                    )

                # spawnu Ref: https://stackoverflow.com/a/37654748/667301
                self.child = px.spawn(
                    self.connect_cmd, timeout=self.login_timeout, encoding=self.encoding, echo=False
                )

            except px.exceptions.EOF as ee:
                time.sleep(70)

            # Ensure the pexpect connection is alive.
            try:
                assert self.child.isalive()

            except AssertionError as ee:
                rich_print(
                    "    [bold red]after_CONNECT_cb()\n        FATAL Assertion: self.child.isalive().\n        AssertionError: %s[/bold red]"
                    % str(ee)
                )
                sys.exit(1)

            # log to screen if requested
            if self.log_screen and self.log_file == "":
                self.child.logfile = sys.stdout

            # log to file if requested
            elif (self.log_screen is False) and self.log_file != "":
                self.child.logfile = open(self.log_file, "w")

            # log to both screen and file if requested
            elif (self.log_screen is True) and self.log_file != "":
                #self.child.logfile = TeeStdoutFile(
                self.child.logfile = TeeStdoutFile(
                    log_file=self.log_file, log_screen=self.log_screen
                )

            if self.debug:
                rich_print(
                    "    [bold blue]after_CONNECT_cb() calling cexpect()[/bold blue]"
                )
            index = self.cexpect(self.base_prompt_regex, timeout=self.login_timeout)

            # Ensure we did not match a UUID prompt...
            assert self.base_prompt_regex[index][0:6] != "UUID~"

            if self.debug:
                rich_print(
                    "    [bold blue]after_CONNECT_cb() call to cexpect() returned index={}[/bold blue]".format(
                        index
                    )
                )

            # Fix ssh cipher and key exchange problems and re-spawn ssh...
            if self.modify_ssh_parameters() is True:
                if self.debug:
                    rich_print(
                        "[bold yellow]modify_ssh_parameters() is closing and reconfiguring ssh[/bold yellow]"
                    )
                self.close()
                self.build_connect_cmd()

                if self.debug:
                    rich_print(
                        "    [bold blue]after_CONNECT_cb() calling itself[/bold blue]"
                    )
                self._go_CONNECT()

            # Do *not* delete or move self.match_prompt; cipher & key exchange detection is there
            # self.matching_prompt

            # FIXME - prompt index should NOT be zero here... need to know why
            if self.debug:
                rich_print(
                    "        [bold blue]self.cexpect() matched prompt index=[/bold blue]{}".format(
                        index
                    )
                )
                rich_print(
                    "        [bold blue]self.cexpect() matching prompt:[/bold blue]'{}'".format(
                        self.base_prompt_regex[index]
                    )
                )

            if index == -1:
                if self.debug:
                    rich_print(
                        "    [bold blue]after_CONNECT_cb() did not match a prompt.[/bold blue]"
                    )
                    rich_print(
                        "    [bold blue]after_CONNECT_cb() is sleeping for 70 seconds.[/bold blue]"
                    )
                time.sleep(70)
                self.build_connect_cmd()
                if self.debug:
                    rich_print(
                        "    [bold blue]after_CONNECT_cb() calling itself[/bold blue]"
                    )
                self._go_CONNECT()

            elif index >= 2:
                # Exit while() loop... ready for username / password...
                ready_for_login = True

        if index == 2:
            if self.debug:
                rich_print("")
                rich_print(
                    "    [bold blue]after_CONNECT_cb() calling after_SEND_LOGIN_USERNAME_cb()[/bold blue]"
                )
            self._go_SEND_LOGIN_USERNAME()
        elif index == 3:
            if self.debug:
                rich_print("")
                rich_print(
                    "    [bold blue]after_CONNECT_cb() calling after_SEND_LOGIN_PASSWORD_cb()[/bold blue]"
                )
            self._go_SEND_LOGIN_PASSWORD()
        elif index == 4:
            if self.debug:
                rich_print("")
                rich_print(
                    "    [bold blue]after_CONNECT_cb() calling after_LOGIN_SUCCESS_UNPRIV_cb()[/bold blue]"
                )
            self._go_LOGIN_SUCCESS_UNPRIV()
        elif index == 5:
            if self.debug:
                rich_print("")
                rich_print(
                    "    [bold blue]after_CONNECT_cb() calling after_LOGIN_SUCCESS_UNPRIV_cb()[/bold blue]"
                )
            self._go_LOGIN_SUCCESS_UNPRIV()
        elif index == 6:
            # Login priv in router prompt=#
            if self.debug:
                rich_print("")
                rich_print(
                    "    [bold blue]after_CONNECT_cb() calling _after_LOGIN_SUCCESS_PRIV_cb()[/bold blue]"
                )
            self._go_LOGIN_SUCCESS_PRIV()
        else:
            raise NotImplementedError()

    def after_SEND_LOGIN_PASSWORD_cb(self):

        if self.debug:
            rich_print("")
            rich_print(
                "[bold blue]Entering state: [/bold blue][bold magenta]{}[/bold magenta]".format(
                    self.state
                )
            )

        self.csendline(self.password)

        try:

            assert len(self.base_prompt_regex) == BASE_PROMPT_REGEX_LENGTH
            if self.debug:
                rich_print(
                    "    [bold blue]after_SEND_LOGIN_PASSWORD_cb() calling cexpect()[/bold blue]"
                )
            index = self.cexpect(self.base_prompt_regex, timeout=self.login_timeout)

            if self.debug:
                rich_print(
                    "\n   [bold blue]after_SEND_LOGIN_PASSWORD_cb() - self.child.expect() matched prompt index:[/bold blue] {}".format(
                        index
                    )
                )
                rich_print(
                    "\n   [bold blue]after_SEND_LOGIN_PASSWORD_cb() - self.child.expect() matching_ prompt:[/bold blue] {}".format(
                        re.escape(self.base_prompt_regex[index])
                    )
                )

            assert index >= 2

            if index == 0:  # This was the wrong ssh cipher...
                if self.debug:
                    rich_print(
                        "        [bold blue]after_SEND_LOGIN_PASSWORD_cb() found index={}.  Closing connection.[/bold blue]".format(
                            index
                        )
                    )
                self.close()
                raise NotImplementedError()

            elif index == 1:  # This was the wrong ssh key exchange...
                if self.debug:
                    rich_print(
                        "        [bold blue]after_SEND_LOGIN_PASSWORD_cb() found index={}.  Closing connection.[/bold blue]".format(
                            index
                        )
                    )
                self.close()
                raise NotImplementedError()

            elif index == 2:
                if self.debug:
                    rich_print(
                        "        [bold blue]after_SEND_LOGIN_PASSWORD_cb() found index={}.  Calling after_SELECT_LOGIN_CREDENTIALS_cb()[/bold blue]".format(
                            index
                        )
                    )
                foo_credentials = (
                    self._go_SELECT_LOGIN_CREDENTIALS()
                )  # Restart login with different creds
                rich_print("[bold red]{}[/bold red]".format(foo_credentials))

                if self.debug:
                    rich_print(
                        "        [bold blue]after_SEND_LOGIN_PASSWORD_cb() found index={}.  Calling after_SEND_LOGIN_USERNAME_cb()[/bold blue]".format(
                            index
                        )
                    )
                self._go_SEND_LOGIN_USERNAME()

            elif index == 3:
                # Send login password again... just in case...
                if self.debug:
                    rich_print(
                        "        [bold blue]after_SEND_LOGIN_PASSWORD_cb() found index={}.  Calling itself with the same password.[/bold blue]".format(
                            index
                        )
                    )
                self._go_SEND_LOGIN_PASSWORD()

            elif index == 4:
                if self.debug:
                    rich_print(
                        "        [bold blue]after_SEND_LOGIN_PASSWORD_cb() found index={}.  Calling after_LOGIN_SUCCESS_UNPRIV_cb()[/bold blue]".format(
                            index
                        )
                    )
                self._go_LOGIN_SUCCESS_UNPRIV()  # We got an unpriv prompt

            elif index == 5:
                if self.debug:
                    rich_print(
                        "        [bold blue]after_SEND_LOGIN_PASSWORD_cb() found index={}.  Calling after_LOGIN_SUCCESS_UNPRIV_cb()[/bold blue]".format(
                            index
                        )
                    )
                self._go_LOGIN_SUCCESS_UNPRIV()  # We got an unpriv prompt

            elif index == 6:
                if self.debug:
                    rich_print(
                        "        [bold blue]after_SEND_LOGIN_PASSWORD_cb() found index={}.  Calling after_LOGIN_SUCCESS_PRIV_cb()[/bold blue]".format(
                            index
                        )
                    )
                self._go_LOGIN_SUCCESS_PRIV()  # We got a priv prompt

            else:
                raise NotImplementedError()

        except px.exceptions.EOF:
            if self.debug:
                rich_print(
                    "   [bold red]after_SEND_LOGIN_PASSWORD_cb() - pexpect.exceptions.EOF error[/bold red]"
                )
            self.close()
            # self._go_SELECT_LOGIN_CREDENTIALS()  # Restart login with different creds
            self._go_CONNECT()  # Restart login with different creds

        except px.exceptions.TIMEOUT:
            if self.debug:
                rich_print(
                    "   [bold red]after_SEND_LOGIN_PASSWORD_cb() - pexpect error: TIMEOUT[/bold red]"
                )
            self.close()
            # self._go_SELECT_LOGIN_CREDENTIALS()  # Restart login with different creds
            if self.debug:
                rich_print(
                    "    [bold blue]after_SEND_LOGIN_PASSWORD_cb() calling after_CONNECT_cb()[/bold blue]"
                )
            self._go_CONNECT()  # Restart login with different creds

    def after_SEND_LOGIN_USERNAME_cb(self):

        if self.debug:
            rich_print("")
            rich_print(
                "[bold blue]Entering state: [/bold blue][bold magenta]{}[/bold magenta]".format(
                    self.state
                )
            )

        self.csendline(self.username)

        if self.debug:
            rich_print(
                "    [bold blue]after_SEND_LOGIN_USERNAME_cb() calling cexpect()[/bold blue]"
            )
        index = self.cexpect(self.base_prompt_regex, timeout=self.login_timeout)

        if self.debug:
            rich_print(
                "   [bold blue]after_SEND_LOGIN_USERNAME_cb() - self.child.expect() matched prompt index:[/bold blue] {}".format(
                    index
                )
            )
            rich_print(
                "   [bold blue]after_SEND_LOGIN_USERNAME_cb() - self.child.expect() matching_prompt:[/bold blue] {}".format(
                    self.base_prompt_regex[index]
                )
            )

        assert index >= 2  # We at least got another username prompt...
        # if index==0:
        #    self._go_SELECT_LOGIN_CREDENTIALS()

        if index == 2:
            self._go_SEND_LOGIN_USERNAME()

        elif index == 3:
            self._go_SEND_LOGIN_PASSWORD()

        elif index == 4:
            self._go_LOGIN_SUCCESS_UNPRIV()  # We got an unpriv prompt here

        elif index == 5:
            self._go_LOGIN_SUCCESS_UNPRIV()  # We got an unpriv prompt here

        elif index == 6:
            self._go_LOGIN_SUCCESS_PRIV()  # We got a priv prompt here

        else:
            raise NotImplementedError()

    def after_LOGIN_SUCCESS_UNPRIV_cb(self):

        if self.debug:
            rich_print("")
            rich_print(
                "[bold blue]Entering state: [/bold blue][bold magenta]{}[/bold magenta]".format(
                    self.state
                )
            )

        if self.debug:
            rich_print(
                "    [bold blue]after_LOGIN_SUCCESS_UNPRIV_cb() calling detect_prompt()[/bold blue]"
            )
        self.detect_prompt()

        if False:
            if self.debug and False:
                rich_print(
                    "   [bold blue]after_LOGIN_SUCCESS_UNPRIV_cb() - self.child.expect() matched prompt index:[/bold blue] {}".format(
                        index
                    )
                )
                rich_print(
                    "   [bold blue]after_LOGIN_SUCCESS_UNPRIV_cb() - self.child.expect() matching_prompt:[/bold blue] {}".format(
                        self.base_prompt_regex[index]
                    )
                )

            assert index >= 3
            if index == 3:
                if self.debug:
                    rich_print(
                        "    [bold blue]after_LOGIN_SUCCESS_UNPRIV_cb() calling after_SEND_PRIV_PASSWORD_cb()[/bold blue]"
                    )
                # Got a 'assword:' prompt
                self._go_SEND_PRIV_PASSWORD()

            elif index == 4:
                # Got a '>' prompt
                if self.debug:
                    rich_print(
                        "    [bold blue]after_LOGIN_SUCCESS_UNPRIV_cb() calling after_LOGIN_COMPLETE_cb()[/bold blue]"
                    )
                self._go_LOGIN_COMPLETE()  # FIXME - is this the right call?

            elif index == 5:
                # Got a '$' prompt
                if self.debug:
                    rich_print(
                        "    [bold red]after_LOGIN_SUCCESS_UNPRIV_cb() hit a fatal exception[/bold red]"
                    )
                raise Exception("index=5 Don't know what to do here")

            elif index == 6:
                if self.debug:
                    rich_print(
                        "    [bold blue]after_LOGIN_SUCCESS_UNPRIV_cb() calling after_LOGIN_SUCCESS_PRIV_cb()[/bold blue]"
                    )
                self._go_LOGIN_SUCCESS_PRIV()  # We got a priv prompt here

            else:
                raise NotImplementedError()

        if self.debug:
            rich_print(
                "    [bold blue]after_LOGIN_SUCCESS_UNPRIV_cb() calling after_LOGIN_COMPLETE_cb()[/bold blue]"
            )
        self._go_LOGIN_COMPLETE()

    def after_SEND_PRIV_PASSWORD_cb(self):

        if self.debug:
            rich_print("")
            rich_print(
                "[bold blue]Entering state: [/bold blue][bold magenta]{}[/bold magenta]".format(
                    self.state
                )
            )

        self.csendline(self.password)

        if self.debug:
            rich_print(
                "    [bold blue]after_SEND_PRIV_PASSWORD_cb() calling cexpect()[/bold blue]"
            )
        index = self.cexpect(self.base_prompt_regex, timeout=self.command_timeout)

        if self.debug:
            rich_print(
                "   [bold blue]after_SEND_PRIV_PASSWORD_cb() - self.child.expect() matched prompt index:[/bold blue] {}".format(
                    index
                )
            )
            rich_print(
                "   [bold blue]after_SEND_PRIV_PASSWORD_cb() - self.child.expect() matching_prompt:[/bold blue] {}".format(
                    self.base_prompt_regex[index]
                )
            )

        if index == 0:
            raise UnexpectedPrompt("Unexpected prompt: 'name:'")
        elif index == 1:
            self._go_SEND_PRIV_PASSWORD()
        elif index == 2:
            raise UnexpectedPrompt("Unexpected prompt: '>'")
        elif index == 3:
            raise UnexpectedPrompt("Unexpected prompt: '$'")
        elif index == 4:
            self._go_LOGIN_SUCCESS_PRIV()  # We got a priv prompt here

    def after_LOGIN_SUCCESS_PRIV_cb(self):

        if self.debug:
            rich_print("")
            rich_print(
                "[bold blue]Entering state: [/bold blue][bold magenta]{}[/bold magenta]".format(
                    self.state
                )
            )

        if self.prompt_hostname == "":
            if self.debug:
                rich_print("    [bold blue]self.prompt_hostname is empty.[/bold blue]")
                rich_print(
                    "    [bold blue]after_LOGIN_COMPLETE_cb() calling detect_prompt()[/bold blue]"
                )
            self.detect_prompt()  # detect_prompt() *should* come before sync_prompt()

        self._go_LOGIN_COMPLETE()

    def after_LOGIN_COMPLETE_cb(self):
        """Clean up on any queued command prompts and accept user commands"""
        # Ensure we don't try to cycles through any more passwords...
        self.login_attempts = 0

        if self.debug:
            rich_print("")
            rich_print(
                "[bold blue]Entering state: [/bold blue][bold magenta]{}[/bold magenta]".format(
                    self.state
                )
            )

        assert self.prompt_hostname != ""

    def after_INTERACT_cb(self):
        """Allow unscripted interaction with the system"""

        if self.debug:
            rich_print("")
            rich_print(
                "[bold blue]Entering state: [/bold blue][bold magenta]{}[/bold magenta]".format(
                    self.state
                )
            )

        assert self.child.isalive()
        rich_print(
            "[bold magenta]Enter INTERACT mode; use Cntl-] to escape[/bold magenta]"
        )

        # FIXME - build a proper output_filter
        self.child.interact(
            escape_character=chr(4), input_filter=None, output_filter=self.interact_output_filter
        )


if __name__ == "__main__":
    sess = Shell(
        host="route-views.oregon-ix.net",
        credentials=(Account(user="rviews", passwd=""),),
        log_screen=False,
        auto_priv_mode=False,
        debug=True,
    )
    sess.execute("term len 0")
    values = sess.execute(
        "show ip int brief",
        template="""Value INTF (\S+)\nValue IPADDR (\S+)\nValue STATUS (up|down|administratively down)\nValue PROTO (up|down)\n\nStart\n  ^${INTF}\s+${IPADDR}\s+\w+\s+\w+\s+${STATUS}\s+${PROTO} -> Record""",
    )
    print("VALUES " + str(values))
    sess.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cannon']

package_data = \
{'': ['*']}

install_requires = \
['arrow==1.2.1',
 'exscript==2.6.3',
 'loguru==0.5.3',
 'textfsm==1.1.2',
 'traits==6.3.2']

setup_kwargs = {
    'name': 'cannon',
    'version': '0.0.69',
    'description': 'An SSH automation tool based on Exscript',
    'long_description': '# Introduction\n\n[cannon][1] is a wrapper around [exscript][2] to connect with remote server or network \ndevices with ssh.\n\n\n## Example Usage - Cisco IOS\n\nThis script will login, run a few show commands.  If you want an interactive session, set `interact=True` when calling Shell()\n\n```python\nimport sys\n\nfrom cannon import Shell, Account\nfrom loguru import logger\n\nlog_stderr_id = logger.add(sink=sys.stderr)\n\n@logger.catch(default=True, onerror=lambda _: sys.exit(1))\ndef main():\n    sess = Shell(\n        host=\'route-views.routeviews.org\',\n        # route-views doesn\'t need password\n        account= Account(name=\'rviews\', password=\'\'),\n        debug=0,\n        json_logfile=\'/tmp/cmd_log.json\',\n        )\n\n    sess.execute(\'term len 0\')\n\n    sess.execute(\'show clock\')\n\n    sess.execute(\'show version\')\n    version_text = sess.response\n\n    # template is a TextFSM template\n    values = sess.execute(\'show ip int brief\',\n        template="""Value INTF (\\S+)\\nValue IPADDR (\\S+)\\nValue STATUS (up|down|administratively down)\\nValue PROTO (up|down)\\n\\nStart\\n  ^${INTF}\\s+${IPADDR}\\s+\\w+\\s+\\w+\\s+${STATUS}\\s+${PROTO} -> Record""")\n    print("VALUES "+str(values))\n    sess.close()\n```\n\n## Example Usage - Linux\n\n```python\nfrom getpass import getpass\nimport sys\n\nfrom cannon.main import Shell, Account\n\nlog_stderr_id = logger.add(sink=sys.stderr)\n\n@logger.catch(default=True, onerror=lambda _: sys.exit(1))\ndef main():\n    account = Account("mpenning", getpass("Login password: "))\n    conn = Shell(host="127.0.0.1", port=22, account=account, driver="generic", debug=0)\n    assert conn is not None\n    example_tfsm_template = """Value UNAME_LINE (.+)\n\nStart\n  ^${UNAME_LINE}\n"""\n    print(conn.execute("sudo uname -a", debug=0, template=example_tfsm_template, timeout=2))\n    print(conn.execute("whoami", debug=0, template=None, timeout=2))\n    #print("FOO2", conn.response)\n    conn.close(force=True)\n\nif __name__=="__main__":\n    main()\n```\n\n## Example test suite setup\n\n- `git clone git@github.com:knipknap/Exscript`\n- `cd` into `Exscript/tests/Exscript/protocols` and `chmod 600 id_rsa`\n- exscript spawns a local tests ssh daemon, `pytest Exscript/tests/Exscript/protocols/SSH2Test.py`\n- Connect with `ssh -i id_rsa -p 1236 user@localhost`\n- one command is supported: `ls`\n\n  [1]: https://pypi.python.org/pypi/cannon    # cannon on pypi\n  [2]: https://pypi.python.org/pypi/exscript  # Exscript on pypi\n',
    'author': 'Mike Pennington',
    'author_email': 'mike@pennington.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mpenning/cannon',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)
